"""A virtual tile that ensapsulates resuable behavior."""

from future.utils import itervalues
import os
import imp
import inspect
import pkg_resources
from iotile.core.exceptions import ExternalError, ArgumentError
from .base_runnable import BaseRunnable
from .common_types import tile_rpc, RPCDispatcher


class VirtualTile(BaseRunnable, RPCDispatcher):
    """A virtual tile.

    Tiles have their own RPC based API and can run background
    threads to do periodic work.

    Args:
        address (int): The address of this tile in the VirtualIOTIleDevice
            that contains it
        name (str): The 6 character name that should be returned when this
            tile is asked for its status to allow matching it with a proxy
            object.
        device (TileBasedVirtualDevice) : optional, device on which this tile is running
    """

    def __init__(self, address, name, device=None):
        super(VirtualTile, self).__init__()

        self.address = address
        self.name = self._check_convert_name(name)

    @classmethod
    def _check_convert_name(cls, name):
        if not isinstance(name, bytes):
            name = name.encode('utf-8')
        if len(name) < 6:
            name += b' '*(6 - len(name))
        elif len(name) > 6:
            raise ArgumentError("Virtual tile name is too long, it must be 6 or fewer characters")

        return name

    def start(self, channel=None):
        """Start any background workers on this tile."""
        self.start_workers()

    def stop(self):
        """Stop any background workers on this tile."""
        self.stop_workers()

    def signal_stop(self):
        """Asynchronously signal that all workers should stop."""
        self.stop_workers_async()

    def wait_stopped(self):
        """Wait for all workers to stop."""
        self.wait_workers_stopped()

    @classmethod
    def FindByName(cls, name):
        """Find an installed VirtualTile by name.

        This function searches for installed virtual tiles
        using the pkg_resources entry_point `iotile.virtual_tile`.

        If name is a path ending in .py, it is assumed to point to
        a module on disk and loaded directly rather than using
        pkg_resources.

        Args:
            name (str): The name of the tile to search
                for.

        Returns:
            VirtualTile class: A virtual tile subclass that can be
                instantiated to create a virtual tile.
        """

        if name.endswith('.py'):
            return cls.LoadFromFile(name)

        for entry in pkg_resources.iter_entry_points("iotile.virtual_tile", name):
            obj = entry.load()
            if not issubclass(obj, VirtualTile):
                raise ExternalError("External virtual tile could not be loaded because it does not inherit from VirtualTile")

            return obj

        raise ArgumentError("VirtualTile could not be found by name", name=name)

    @classmethod
    def LoadFromFile(cls, script_path):
        """Import a virtual tile from a file rather than an installed module

        script_path must point to a python file ending in .py that contains exactly one
        VirtualTile class definition.  That class is loaded and executed as if it
        were installed.

        To facilitate development, if there is a proxy object defined in the same
        file, it is also added to the HardwareManager proxy registry so that it
        can be found and used with the device.

        Args:
            script_path (string): The path to the script to load

        Returns:
            VirtualTile: A subclass of VirtualTile that was loaded from script_path
        """

        search_dir, filename = os.path.split(script_path)
        if search_dir == '':
            search_dir = './'

        if filename == '' or not os.path.exists(script_path):
            raise ArgumentError("Could not find script to load virtual tile", path=script_path)

        module_name, ext = os.path.splitext(filename)
        if ext != '.py':
            raise ArgumentError("Script did not end with .py", filename=filename)

        try:
            file_obj = None
            file_obj, pathname, desc = imp.find_module(module_name, [search_dir])
            mod = imp.load_module(module_name, file_obj, pathname, desc)
        finally:
            if file_obj is not None:
                file_obj.close()

        devs = [x for x in itervalues(mod.__dict__) if inspect.isclass(x) and issubclass(x, VirtualTile) and x != VirtualTile]
        if len(devs) == 0:
            raise ArgumentError("No VirtualTiles subclasses were defined in script", path=script_path)
        elif len(devs) > 1:
            raise ArgumentError("More than one VirtualTiles subclass was defined in script", path=script_path, tiles=devs)

        return devs[0]

    @tile_rpc(0x0004, "", "H6sBBBB")
    def tile_status(self):
        """Required status RPC that allows matching a proxy object with a tile."""

        status = (1 << 1) | (1 << 0)  # Configured and running, not currently used but required for compat with physical tiles
        return [0xFFFF, self.name, 1, 0, 0, status]



"""Base class for objects that have a background runnable task like virtual devices and tiles."""

from iotile.core.exceptions import InternalError
from iotile.core.utilities.stoppable_thread import StoppableWorkerThread


class BaseRunnable(object):
    """A simple class that uses StoppableWorkerThreads for background tasks."""

    def __init__(self):
        super(BaseRunnable, self).__init__()

        self._workers = []
        self._started = False

    def create_worker(self, func, interval, *args, **kwargs):
        """Spawn a worker thread running func.

        The worker will be automatically be started when start() is called
        and terminated when stop() is called on this object.
        This must be called only from the main thread, not from a worker thread.

        create_worker must not be called after stop() has been called.  If it
        is called before start() is called, the thread is started when start()
        is called, otherwise it is started immediately.

        Args:
            func (callable): Either a function that will be called in a loop
                with a sleep of interval seconds with *args and **kwargs or
                a generator function that will be called once and expected to
                yield periodically so that the worker can check if it should
                be killed.
            interval (float): The time interval between invocations of func.
                This should not be 0 so that the thread doesn't peg the CPU
                and should be short enough so that the worker checks if it
                should be killed in a timely fashion.
            *args: Arguments that are passed to func as positional args
            **kwargs: Arguments that are passed to func as keyword args
        """

        thread = StoppableWorkerThread(func, interval, args, kwargs)
        self._workers.append(thread)

        if self._started:
            thread.start()

    def start_workers(self):
        """Start running this virtual device including any necessary worker threads."""

        if self._started:
            raise InternalError("The method start() was called twice on a BaseRunnable object.")

        self._started = True

        for worker in self._workers:
            worker.start()

    def stop_workers(self):
        """Synchronously stop any potential workers."""

        self._started = False

        for worker in self._workers:
            worker.stop()

    def stop_workers_async(self):
        """Signal that all workers should stop without waiting."""

        self._started = False
        for worker in self._workers:
            worker.signal_stop()

    def wait_workers_stopped(self):
        """Wait for all workers to stop."""

        for worker in self._workers:
            worker.wait_stopped()



"""A recursive verifier that verifies the schema of a python dictionary

This is similar to the concept of a JSON schema verifier but with more expressive
verification steps and works on dictionaries produced from a variety of sources.
"""

import inspect
from copy import deepcopy
from iotile.core.exceptions import ValidationError

class Verifier(object):
    """A base class for verifing that an object conforms to a schema

    Subclasses should override the verify function to actually verify the
    object that is passed in.

    Args:
        desc (string): An optional block description of what this verifier is
            checking.
    """

    def __init__(self, desc=None):
        self.short_desc = None
        self.long_desc = None
        self.description = None
        if desc is not None:
            self.set_description(desc)

    def clone(self):
        """Clone this verifier.

        This function is useful for implementing hierarch among verifiers
        where one or more of the verification steps is shared.

        For example, a dict that always has key1 and key2 can be pulled
        into a single DictVerifier and cloned for each variant of the dict.
        """

        return deepcopy(self)

    def set_description(self, desc):
        self.description = inspect.cleandoc(desc)

        self.short_desc = self._get_short_description()
        self.long_desc = self._get_long_description()

    def matches(self, obj):
        """Return True if object matches this verifier."""

        try:
            self.verify(obj)
            return True
        except ValidationError:
            return False

    def verify(self, obj):
        """Verify that the object conforms to this verifier's schema.

        Args:
            obj (object): A python object to verify

        Raises:
            ValidationError: If there is a problem verifying the dictionary, a
                ValidationError is thrown with at least the reason key set indicating
                the reason for the lack of validation.
        """

        return obj

    def _get_short_description(self):
        """Return the first line of a multiline description

        Returns:
            string: The short description, otherwise None
        """

        if self.description is None:
            return None

        lines = [x for x in self.description.split('\n')]
        if len(lines) == 1:
            return lines[0]
        elif len(lines) >= 3 and lines[1] == '':
            return lines[0]

        return None

    def _get_long_description(self):
        """Return the subsequent lines of a multiline description

        Returns:
            string: The long description, otherwise None
        """

        if self.description is None:
            return None

        lines = [x for x in self.description.split('\n')]
        if len(lines) == 1:
            return None

        elif len(lines) >= 3 and lines[1] == '':
            return '\n'.join(lines[2:])

        return self.description

    def wrap_lines(self, text, indent_level, indent_size=4):
        """Indent a multiline string

        Args:
            text (string): The string to indent
            indent_level (int): The number of indent_size spaces to prepend
                to each line
            indent_size (int): The number of spaces to prepend for each indent
                level

        Returns:
            string: The indented block of text
        """

        indent = ' '*indent_size*indent_level
        lines = text.split('\n')

        wrapped_lines = []

        for line in lines:
            if line == '':
                wrapped_lines.append(line)
            else:
                wrapped_lines.append(indent + line)

        return '\n'.join(wrapped_lines)

    def format_name(self, name, indent_size=4):
        """Format the name of this verifier

        The name will be formatted as:
            <name>: <short description>
                long description if one is given followed by \n
                otherwise no long description

        Args:
            name (string): A name for this validator
            indent_size (int): The number of spaces to indent the
                description
        Returns:
            string: The formatted name block with a short and or long
                description appended.
        """

        name_block = ''

        if self.short_desc is None:
            name_block += name + '\n'
        else:
            name_block += name + ': ' + self.short_desc + '\n'

        if self.long_desc is not None:
            name_block += self.wrap_lines(self.long_desc, 1, indent_size=indent_size)
            name_block += '\n'

        return name_block

    def trim_whitespace(self, text):
        """Remove leading whitespace from each line of a multiline string

        Args:
            text (string): The text to be unindented

        Returns:
            string: The unindented block of text
        """

        lines = text.split('\n')
        new_lines = [x.lstrip() for x in lines]

        return '\n'.join(new_lines)



"""Base classes for creating and interacting with virtual IOTile devices."""

from .virtualtile import VirtualTile
from .virtualdevice import VirtualIOTileDevice

from .common_types import (tile_rpc, RPCDispatcher, RPCInvalidIDError,
                           RPCNotFoundError, RPCInvalidArgumentsError,
                           RPCInvalidReturnValueError, TileNotFoundError,
                           RPCErrorCode, unpack_rpc_payload)
from .virtualinterface import VirtualIOTileInterface

__all__ = ['VirtualTile', 'VirtualIOTileDevice', 'tile_rpc',
           'RPCDispatcher', 'RPCInvalidIDError', 'TileNotFoundError',
           'RPCNotFoundError', 'RPCInvalidArgumentsError',
           'RPCInvalidReturnValueError', 'RPCErrorCode', 'VirtualIOTileInterface', 'unpack_rpc_payload']



# Annotated Registry
# A wrapper to make the IOTile component registry accessible in the iotile
# tool.  Since the registry is used internally in the type system it cannot
# itself make use of typedargs annotations
from iotile.core.utilities.typedargs import annotated, param, return_type, context, iprint

_name_ = "Developer"

#Outside accessible API for this package
from .registry import ComponentRegistry

@annotated
def registry():
    return AnnotatedRegistry()

@context()
class AnnotatedRegistry:
    """
    AnnotatedRegistry

    A mapping of all of the installed components on this system that can
    be used as build dependencies and where they are located.  Also used
    to manage iotile plugins.
    """

    def __init__(self):
        self.reg = ComponentRegistry()

    @param("component", "path", "exists", desc="folder containing component to register")
    def add_component(self, component):
        """
        Register a component with ComponentRegistry.

        Component must be a buildable object with a module_settings.json file that
        describes its name and the domain that it is part of.
        """

        self.reg.add_component(component)

    @return_type("map(string, string)")
    def list_plugins(self):
        """
        List all of the plugins that have been registerd for the iotile program on this computer
        """

        return self.reg.list_plugins()

    def find_component(self, key, domain=""):
        return self.reg.find_component(key, domain)

    @param("key", "string", desc="iotile component to find")
    def remove_component(self, key):
        """
        Remove component from registry
        """

        return self.reg.remove_component(key)

    @annotated
    def clear_components(self):
        """
        Clear all of the registered components
        """

        self.reg.clear_components()

    @return_type("list(string)")
    def list_components(self):
        """
        List all of the registered component names
        """

        return self.reg.list_components()

    @return_type("list(string)")
    def list_config(self):
        """
        List all of the registered component names
        """

        return self.reg.list_config()

    @param("key", "string")
    @param("value", "string")
    def set(self, key, value):
        self.reg.set_config(key, value)

    @param("key", "string")
    @return_type("string")
    def get(self, key):
        return self.reg.get_config(key)

    @param("key", "string")
    @return_type("string")
    def clear(self, key):
        self.reg.clear_config(key)



from __future__ import (unicode_literals, print_function, absolute_import)
from builtins import str
from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.exceptions import ArgumentError

class FlashBoardStep (object):
    """A Recipe Step used to flash firmware directly
    
    Currently only supports using jlink to flash. Used to bootstrap firmware

    This function requires a shared hardware manager resource to be setup
    containing a connected device that we can send the script to.

    Args:
        file (str): Firmware file name to flash
        debug_string (str): Optional connection string for debug
    """
    REQUIRED_RESOURCES = [('connection', 'hardware_manager')]
    FILES = ['file']

    def __init__(self, args):
        if 'file' not in args:
            raise ArgumentError("FlashBoardStep required parameters missing", required=["file"], args=args)

        self._file = args['file']
        self._debug_string = args.get('debug_string')

    def run(self, resources):
        """Runs the flash step

        Args:
            resources (dict): A dictionary containing the required resources that
                we needed access to in order to perform this step.
        """
        if not resources['connection']._port.startswith('jlink'):
            raise ArgumentError("FlashBoardStep is currently only possible through jlink", invalid_port=args['port'])

        hwman = resources['connection']
        debug = hwman.hwman.debug(self._debug_string)
        debug.flash(self._file)


"""A list of event codes that are logged to the audit log by virtual devices for verification purposes
"""

from collections import namedtuple

AuditEvent = namedtuple('AuditEvent', ['name', 'message', 'required_keys', 'optional_keys'])

ClientConnected = AuditEvent('ClientConnected', 'A client connected to this device', frozenset(), frozenset())
ClientDisconnected = AuditEvent('ClientDisconnected', 'A client disconnected from this device', frozenset(), frozenset())
RPCInterfaceOpened = AuditEvent('RPCInterfaceOpened', 'A client opened the RPC interface on this device', frozenset(), frozenset())
StreamingInterfaceOpened = AuditEvent('StreamingInterfaceOpened', 'A client opened the streaming interface on this device', frozenset(), frozenset())
StreamingInterfaceClosed = AuditEvent('StreamingInterfaceClosed', 'A client closed the streaming interface on this device', frozenset(), frozenset())
ReportStreamed = AuditEvent('ReportStreamed', 'The device streamed a report to the client, report=%(report)s', frozenset(['report']), frozenset())
TraceSent = AuditEvent('TraceSent', 'The device streamed tracing data to the client, data=%(trace)s', frozenset(['trace']), frozenset())
ErrorStreamingReport = AuditEvent('ErrorStreamingReport', 'There was an error sending a report to the client, further streaming has been stopped', frozenset(), frozenset())
TracingInterfaceOpened = AuditEvent('TracingInterfaceOpened', 'A client opened the tracing interface on this device', frozenset(), frozenset())
TracingInterfaceClosed = AuditEvent('TracingInterfaceClosed', 'A client closed the tracing interface on this device', frozenset(), frozenset())
ErrorStreamingTrace = AuditEvent('ErrorStreamingTrace', 'There was an error sending tracing data to the client, further tracing has been stopped', frozenset(), frozenset())
RPCReceived = AuditEvent('RPCReceived', 'An RPC has been processed (id=%(rpc_id)s, address=%(address)s, payload="%(payload)s"), status=%(status)d, response="%(response)s"', frozenset(['rpc_id', 'address', 'payload', 'status', 'response']), frozenset())
ErrorSendingRPCResponse = AuditEvent('ErrorSendingRPCResponse', 'There was a permanent error sending and RPC response', frozenset(), frozenset())
ScriptInterfaceOpened = AuditEvent('ScriptInterfaceOpened', 'A client opened the script interface on this device', frozenset(), frozenset())
ScriptInterfaceClosed = AuditEvent('ScriptInterfaceClosed', 'A client closed the script interface on this device', frozenset(), frozenset())



"""A thread that can be stopped
"""

from builtins import range
import threading
import inspect
import traceback
import time
from iotile.core.exceptions import TimeoutExpiredError


class StoppableWorkerThread(threading.Thread):
    """A worker thread that calls a worker function periodically.

    This class takes a single callable function and calls that
    function with *args and **kwargs in a loop with a configurable
    delay between each invocation.  There is a public stop() method
    that will kill the thread after the next time the callable function
    returns.

    If you pass a generator function then the function is called in a loop
    with next() until it raises a StopIteration exception at which point
    the thread is closed.  Exceptions thrown from the generator cause the
    thread to exit immediately.  The stop condition of the thread is checked
    every time the generator yields.
    """

    def __init__(self, routine, timeout=0.1, args=None, kwargs=None):
        self._routine = routine
        self._worker_args = args
        self._worker_kwargs = kwargs
        self._stop_condition = threading.Event()
        self._running = threading.Event()


        # In order to keep the thread responsive, we default to checking for our
        # stop signal every 10 milliseconds.  If the execution timeout is less than
        # 10 ms, we use that, otherwise we round it to a multiple of 10 ms
        #
        # Allow the user to specify no wait time if the function we are calling has
        # a built-in delay, like it's reading from a file or some other source of
        # events.
        if timeout is None:
            self._wait = 0.0
            self._wait_count = 0
        elif timeout <= 0.01:
            self._wait = timeout
            self._wait_count = 1
        else:
            self._wait = 0.01
            self._wait_count = int(timeout / self._wait)

        self._generator = inspect.isgeneratorfunction(routine)


        if self._worker_args is None:
            self._worker_args = []

        if self._worker_kwargs is None:
            self._worker_kwargs = {}

        super(StoppableWorkerThread, self).__init__()

        self.daemon = True

    @property
    def generator(self):
        """Whether this thread is running a generator function or not
        """

        return self._generator

    def run(self):
        """The target routine called to start thread activity.

        If the thread is created with a generator function, this iterates
        the generator and checks for a stop condition between each iteration.

        If the thread is created with a normal function, that function is called
        in a loop with the stop condition checked between each invocation.
        """

        if self._generator:
            try:
                gen = self._routine(*self._worker_args, **self._worker_kwargs)

                while True:
                    if self._stop_condition.is_set():
                        return

                    self._running.set()

                    next(gen)

                    for _i in range(0, self._wait_count):
                        if self._stop_condition.is_set():
                            return
                        time.sleep(self._wait)

            except StopIteration:
                pass
            except Exception:
                print("Exception occurred in background worker thread")
                traceback.print_exc()
        else:
            try:
                while True:
                    if self._stop_condition.is_set():
                        break

                    self._running.set()

                    self._routine(*self._worker_args, **self._worker_kwargs)

                    # Wait for the desired interval, checking if we should exit
                    for _i in range(0, self._wait_count):
                        if self._stop_condition.is_set():
                            return

                        time.sleep(self._wait)
            except Exception:
                print("Exception occurred in background worker thread")
                traceback.print_exc()

    def wait_running(self, timeout=None):
        """Wait for the thread to pass control to its routine.

        Args:
            timeout (float): The maximum amount of time to wait
        """

        flag = self._running.wait(timeout)

        if flag is False:
            raise TimeoutExpiredError("Timeout waiting for thread to start running")

    def stop(self, timeout=None, force=False):
        """Stop the worker thread and synchronously wait for it to finish.

        Args:
            timeout (float): The maximum time to wait for the thread to stop
                before raising a TimeoutExpiredError.  If force is True, TimeoutExpiredError
                is not raised and the thread is just marked as a daemon thread
                so that it does not block cleanly exiting the process.
            force (bool): If true and the thread does not exit in timeout seconds
                no error is raised since the thread is marked as daemon and will
                be killed when the process exits.
        """

        self.signal_stop()
        self.wait_stopped(timeout, force)

    def signal_stop(self):
        """Signal that the worker thread should stop but don't wait.

        This function is useful for stopping multiple threads in parallel when
        combined with wait_stopped().
        """

        self._stop_condition.set()

    def wait_stopped(self, timeout=None, force=False):
        """Wait for the thread to stop.

        You must have previously called signal_stop or this function will hang.

        Args:
            timeout (float): The maximum time to wait for the thread to stop
                before raising a TimeoutExpiredError.  If force is True, TimeoutExpiredError
                is not raised and the thread is just marked as a daemon thread
                so that it does not block cleanly exiting the process.
            force (bool): If true and the thread does not exit in timeout seconds
                no error is raised since the thread is marked as daemon and will
                be killed when the process exits.
        """

        self.join(timeout)

        if self.is_alive() and force is False:
            raise TimeoutExpiredError("Error waiting for background thread to exit", timeout=timeout)



"""A websocket client that validates messages received and dispatches them."""

from __future__ import unicode_literals
from ws4py.client.threadedclient import WebSocketClient
from future.utils import viewitems
import threading
import msgpack
import datetime
import logging
import uuid
from iotile.core.exceptions import IOTileException, InternalError, ValidationError, TimeoutExpiredError
from iotile.core.utilities.schema_verify import Verifier, DictionaryVerifier, StringVerifier, LiteralVerifier, OptionsVerifier

# The prescribed schema of command response messages
# Messages with this format are automatically processed inside the ValidatingWSClient
SuccessfulResponseSchema = DictionaryVerifier()
SuccessfulResponseSchema.add_required('type', LiteralVerifier('response'))
SuccessfulResponseSchema.add_required('success', LiteralVerifier(True))
SuccessfulResponseSchema.add_optional('payload', Verifier())

FailureResponseSchema = DictionaryVerifier()
FailureResponseSchema.add_required('type', LiteralVerifier('response'))
FailureResponseSchema.add_required('success', LiteralVerifier(False))
FailureResponseSchema.add_required('reason', StringVerifier())

ResponseSchema = OptionsVerifier(SuccessfulResponseSchema, FailureResponseSchema)


class ValidatingWSClient(WebSocketClient):
    """A threaded websocket client that validates messages received.

    Messages are assumed to be packed using msgpack in a binary format
    and are decoded and validated against message type schema.  Matching
    messages are dispatched to the appropriate handler and messages that
    match no attached schema are logged and dropped.
    """

    def __init__(self, url, logger_name=__name__):
        """Constructor.

        Args:
        url (string): The URL of the websocket server that we want
            to connect to.
        logger_name (string): An optional name that errors are logged to
        """

        super(ValidatingWSClient, self).__init__(url)

        self._connected = threading.Event()
        self._disconnection_finished = threading.Event()

        self._command_lock = threading.Lock()

        self.control_data = str(uuid.uuid4())
        self._pong_received = threading.Event()

        self._last_response = None
        self._response_received = threading.Event()

        self.disconnection_code = None
        self.disconnection_reason = None
        self.disconnection_callback = None

        self.logger = logging.getLogger(logger_name)
        self.logger.addHandler(logging.NullHandler())

        self.validators = [(ResponseSchema, self._on_response_received)]

    @property
    def connected(self):
        return self._connected.is_set()

    def add_message_type(self, validator, callback):
        """Add a message type that should trigger a callback.

        Each validator must be unique, in that a message will
        be dispatched to the first callback whose validator
        matches the message.

        Args:
            validator (Verifier): A schema verifier that will
                validate a received message uniquely
            callback (callable): The function that should be called
                when a message that matches validator is received.
        """

        self.validators.append((validator, callback))

    def start(self, timeout=5.0):
        """Synchronously connect to websocket server.

        Args:
            timeout (float): The maximum amount of time to wait for the
                connection to be established. Defaults to 5 seconds
        """

        try:
            self.connect()
        except Exception as exc:
            raise InternalError("Unable to connect to websockets host", reason=str(exc))

        flag = self._connected.wait(timeout=timeout)
        if not flag:
            raise TimeoutExpiredError("Conection attempt to host timed out")

    def stop(self, timeout=5.0):
        """Synchronously disconnect from websocket server.

        Args:
            timeout (float): The maximum amount of time to wait for the
                connection to be established. Defaults to 5 seconds
        """

        if not self.connected:
            return

        try:
            self.close()
        except Exception as exc:
            raise InternalError("Unable to disconnect from websockets host", reason=str(exc))

        flag = self._disconnection_finished.wait(timeout=timeout)
        if not flag:
            raise TimeoutExpiredError("Disconnection attempt from host timed out")
        self._disconnection_finished.clear()

    def send_message(self, obj):
        """Send a packed message.

        Args:
            obj (dict): The message to be sent
        """

        packed = msgpack.packb(obj, use_bin_type=True)
        self.send(packed, binary=True)

    def send_command(self, command, args, timeout=10.0):
        """Send a command and synchronously wait for a response.

        Args:
            command (string): The command name
            args (dict): Optional arguments
            timeout (float): The maximum time to wait for a response
        """

        msg = {x: y for x, y in viewitems(args)}
        msg['type'] = 'command'
        msg['operation'] = command
        msg['no_response'] = False

        with self._command_lock:
            self._response_received.clear()
            self.send_message(msg)

            flag = self._response_received.wait(timeout=timeout)
            if not flag:
                raise TimeoutExpiredError("Timeout waiting for response")

            self._response_received.clear()
            return self._last_response

    def send_ping(self, timeout=10.0):
        """Send a ping message to keep connection alive and to verify
        if the server is still there."""

        self.ping(self.control_data)

        flag = self._pong_received.wait(timeout=timeout)
        if not flag:
            raise TimeoutExpiredError("Timeout waiting for pong response")

        self._pong_received.clear()

    def ponged(self, pong):
        """Pong message received after a ping was sent"""

        if str(pong) == self.control_data:
            self._pong_received.set()

    def post_command(self, command, args):
        """Post a command asynchronously and don't wait for a response.

        Args:
            command (string): The command name
            args (dict): Optional arguments
        """

        msg = {x: y for x, y in viewitems(args)}
        msg['type'] = 'command'
        msg['operation'] = command
        msg['no_response'] = True

        self.send_message(msg)

    def opened(self):
        """Callback called in another thread when a connection is opened."""

        self._connected.set()

    def closed(self, code, reason):
        """Callback called in another thread when a connection is closed.

        Args:
            code (int): A code for the closure
            reason (string): A reason for the closure
        """

        self.disconnection_code = code
        self.disconnection_reason = reason
        self._disconnection_finished.set()
        self._connected.clear()

        if self.disconnection_callback is not None:
            self.disconnection_callback()

    def _unpack(self, msg):
        """Unpack a binary msgpacked message."""

        return msgpack.unpackb(msg, raw=False, object_hook=self.decode_datetime)

    @classmethod
    def decode_datetime(cls, obj):
        """Decode a msgpack'ed datetime."""
        if b'__datetime__' in obj:
            obj = datetime.datetime.strptime(obj[b'as_str'].decode(), "%Y%m%dT%H:%M:%S.%f")
        return obj

    @classmethod
    def encode_datetime(cls, obj):
        """Encode a msgpck'ed datetime."""
        if isinstance(obj, datetime.datetime):
            obj = {'__datetime__': True, 'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f").encode()}
        return obj

    def received_message(self, message):
        """Callback when a message is received.

        The message must be encoded using message_mack

        Args:
            message (object): The message that was received
        """

        try:
            unpacked = self._unpack(message.data)
        except Exception as exc:
            self.logger.error("Corrupt message received, parse exception = %s", str(exc))
            return

        # Look for the first callback that can handle this message
        # if no one can handle it, log an error and discard the message.
        for validator, callback in self.validators:
            try:
                validator.verify(unpacked)
            except ValidationError:
                continue

            try:
                callback(unpacked)
                return
            except IOTileException as exc:
                self.logger.error("Exception handling websocket message, exception = %s", str(exc))
                return
            except Exception as exc:
                self.logger.error("Non-IOTile exception handling websocket message, exception = %s", str(exc))
                return

        self.logger.warn("No handler found for received message, message=%s", str(unpacked))

    def _on_response_received(self, resp):
        self._last_response = resp
        self._response_received.set()



"""An adapter class that takes a DeviceAdapter and produces a CMDStream compatible interface
"""

from builtins import str
from future.utils import viewitems
from copy import deepcopy
import queue
from .cmdstream import CMDStream
import datetime
import time
from iotile.core.exceptions import HardwareError, ArgumentError
from iotile.core.utilities.typedargs import iprint
from iotile.core.hw.reports import BroadcastReport


class AdapterCMDStream(CMDStream):
    """An adapter class that takes a DeviceAdapter and produces a CMDStream compatible interface

    DeviceAdapters have a more generic interface that is not restricted to exclusive use in an online
    fashion by a single user at a time.  This class implements the CMDStream interface on top of a
    DeviceAdapter.

    Args:
        adapter (DeviceAdapter): the DeviceAdatper that we should use to create this CMDStream
        port (string): the name of the port that we should connect through
        connection_string (string): A DeviceAdapter specific string specifying a device that we
            should immediately connect to
        record (string): The path to a file that we should use to record everything sent down
            this CMDStream
    """

    def __init__(self, adapter, port, connection_string, record=None):
        self.adapter = adapter
        self._scanned_devices = {}
        self._reports = None
        self._broadcast_reports = None
        self._traces = None
        self.connection_interrupted = False

        self.adapter.add_callback('on_scan', self._on_scan)
        self.adapter.add_callback('on_report', self._on_report)
        self.adapter.add_callback('on_trace', self._on_trace)
        self.adapter.add_callback('on_disconnect', self._on_disconnect)

        self.start_time = time.time()
        self.min_scan = self.adapter.get_config('minimum_scan_time', 0.0)
        self.probe_required = self.adapter.get_config('probe_required', False)

        super(AdapterCMDStream, self).__init__(port, connection_string, record)

    def _on_scan(self, adapter_id, info, expiration_time):
        """Callback called when a new device is discovered on this CMDStream

        Args:
            adapter_id (int): An ID for the adapter that scanned the device
            info (dict): Information about the scanned device
            expiration_time (float): How long this device should stay around
        """

        device_id = info['uuid']
        infocopy = deepcopy(info)

        infocopy['expiration_time'] = datetime.datetime.now() + datetime.timedelta(seconds=expiration_time)
        self._scanned_devices[device_id] = infocopy

    def _on_disconnect(self, adapter_id, connection_id):
        """Callback when a device is disconnected unexpectedly.

        Args:
            adapter_id (int): An ID for the adapter that was connected to the device
            connection_id (int): An ID for the connection that has become disconnected
        """

        self.connection_interrupted = True

    def _scan(self, wait=None):
        """Return the devices that have been found for this device adapter.

        If the adapter indicates that we need to explicitly tell it to probe for devices, probe now.
        By default we return the list of seen devices immediately, however there are two cases where
        we will sleep here for a fixed period of time to let devices show up in our result list:

        - If we are probing then we wait for 'minimum_scan_time'
        - If we are told an explicit wait time that overrides everything and we wait that long
        """

        # Figure out how long and if we need to wait before returning our scan results
        wait_time = None
        elapsed = time.time() - self.start_time
        if elapsed < self.min_scan:
            wait_time = self.min_scan - elapsed

        # If we need to probe for devices rather than letting them just bubble up, start the probe
        # and then use our min_scan_time to wait for them to arrive via the normal _on_scan event
        if self.probe_required:
            self.adapter.probe_sync()
            wait_time = self.min_scan

        # If an explicit wait is specified that overrides everything else
        if wait is not None:
            wait_time = wait

        if wait_time is not None:
            time.sleep(wait_time)

        to_remove = set()

        now = datetime.datetime.now()

        for name, value in viewitems(self._scanned_devices):
            if value['expiration_time'] < now:
                to_remove.add(name)

        for name in to_remove:
            del self._scanned_devices[name]

        return self._scanned_devices.values()

    def _connect(self, uuid_value, wait=None):
        # If we can't see the device, scan to try to find it
        if uuid_value not in self._scanned_devices:
            self.scan(wait=wait)

        if uuid_value not in self._scanned_devices:
            raise HardwareError("Could not find device to connect to by UUID", uuid=uuid_value)

        connstring = self._scanned_devices[uuid_value]['connection_string']
        self._connect_direct(connstring)

        return connstring

    def _connect_direct(self, connection_string):
        res = self.adapter.connect_sync(0, connection_string)
        if not res['success']:
            self.adapter.periodic_callback()
            raise HardwareError("Could not connect to device", reason=res['failure_reason'], connection_string=connection_string)

        try:
            res = self.adapter.open_interface_sync(0, 'rpc')
        except Exception as exc:
            self.adapter.disconnect_sync(0)
            self.adapter.periodic_callback()
            raise HardwareError("Could not open RPC interface on device due to an exception", exception=str(exc))

        if not res['success']:
            self.adapter.disconnect_sync(0)
            self.adapter.periodic_callback()
            raise HardwareError("Could not open RPC interface on device", reason=res['failure_reason'], connection_string=connection_string)

    def _disconnect(self):

        # Close the streaming and tracing interfaces when we disconnect
        self._reports = None
        self._traces = None

        self.adapter.disconnect_sync(0)
        self.adapter.periodic_callback()

    def _try_reconnect(self):
        """Try to recover an interrupted connection."""

        try:
            if self.connection_interrupted:
                self._connect_direct(self.connection_string)
                self.connection_interrupted = False
                self.connected = True

                # Reenable streaming interface if that was open before as well
                if self._reports is not None:
                    res = self.adapter.open_interface_sync(0, 'streaming')
                    if not res['success']:
                        raise HardwareError("Could not open streaming interface to device", reason=res['failure_reason'])

                # Reenable tracing interface if that was open before as well
                if self._traces is not None:
                    res = self.adapter.open_interface_sync(0, 'tracing')
                    if not res['success']:
                        raise HardwareError("Could not open tracing interface to device", reason=res['failure_reason'])
        except HardwareError as exc:
            raise HardwareError("Device disconnected unexpectedly and we could not reconnect", reconnect_error=exc)

    def _send_rpc(self, address, rpc_id, payload, **kwargs):
        timeout = 3.0
        if 'timeout' in kwargs:
            timeout = float(kwargs['timeout'])

        # If our connection was interrupted before this RPC, try to recover it
        if self.connection_interrupted:
            self._try_reconnect()

        result = self.adapter.send_rpc_sync(0, address, rpc_id, payload, timeout)
        success = result['success']
        status = result['status']
        payload = result['payload']

        # Sometimes RPCs can cause the device to go offline, so try to reconnect to it.
        # For example, the RPC could cause the device to reset itself.
        if self.connection_interrupted:
            self._try_reconnect()

        if not success:
            raise HardwareError("Could not send RPC", reason=result['failure_reason'])

        self.adapter.periodic_callback()

        return status, payload

    def _send_highspeed(self, data, progress_callback):
        if isinstance(data, str) and not isinstance(data, bytes):
            raise ArgumentError("You must send bytes or bytearray to _send_highspeed", type=type(data))

        if not isinstance(data, bytes):
            data = bytes(data)

        self.adapter.send_script_sync(0, data, progress_callback)

    def _enable_streaming(self):
        self._reports = queue.Queue()
        res = self.adapter.open_interface_sync(0, 'streaming')
        if not res['success']:
            raise HardwareError("Could not open streaming interface to device", reason=res['failure_reason'])

        return self._reports

    def _enable_broadcasting(self):
        if self._broadcast_reports is not None:
            return

        self._broadcast_reports = queue.Queue()
        return self._broadcast_reports

    def _enable_debug(self, connection_string=None):
        res = self.adapter.open_interface_sync(0, 'debug', connection_string)
        if not res['success']:
            raise HardwareError("Could not open debug interface to device", reason=res['failure_reason'])

    def _debug_command(self, cmd, args, progress_callback=None):
        def _progress_callback(_finished, _total):
            pass

        res = self.adapter.debug_sync(0, cmd, args, progress_callback)
        if not res['success']:
            raise HardwareError("Could not execute debug command %s on device" % cmd, reason=res['failure_reason'])

        return res.get('return_value')

    def _enable_tracing(self):
        self._traces = queue.Queue()
        res = self.adapter.open_interface_sync(0, 'tracing')
        if not res['success']:
            raise HardwareError("Could not open tracing interface to device", reason=res['failure_reason'])

        return self._traces

    def _on_report(self, conn_id, report):
        if isinstance(report, BroadcastReport):
            if self._broadcast_reports is None:
                return

            self._broadcast_reports.put(report)
            return

        if self._reports is None:
            return

        self._reports.put(report)

    def _on_trace(self, conn_id, tracing_data):
        if self._traces is None:
            return

        self._traces.put(tracing_data)

    def _close(self):
        self.adapter.stop_sync()



"""Helper module for assigning UTC times to readings without a realtime clock.

Some IOTile devices timestamp their readings in uptime rather than UTC time,
which means a conversion process is needed to convert to absolute time from
the device's local reference time.

The UTCAssigner class provides a reference implementation of this conversion
process.  The conversion algorithm looks for known points in time where both
the device uptime and absolute UTC time are known.  These points are known as
anchors.  For any other reading, the difference in uptime between it and the
closest anchor is added to the absolute UTC time of the anchor to calculate
the absolute UTC time of the reading.

In the easy case, when there are no device resets between the reading you want
to convert and an anchor point and uptimes are known for each point, this
reconstruction process is exact.

However, in the general case, an exact assignment is not possible and
UTCAssigner uses various other methods to infer an appoximate UTC timestamp,
returning confidence metrics along with the assigned value.
"""

import datetime
from typedargs.exceptions import ArgumentError
from sortedcontainers import SortedKeyList
from .signed_list_format import SignedListReport


class _TimeAnchor(object):
    """Internal class for storing a utc reference point."""

    __slots__ = ('uptime', 'utc', 'reading_id', 'is_break')

    def __init__(self, reading_id, uptime=None, utc=None, is_break=False):
        self.uptime = uptime
        self.utc = utc
        self.reading_id = reading_id
        self.is_break = is_break

    def copy(self):
        return _TimeAnchor(self.reading_id, self.uptime, self.utc, self.is_break)


class UTCAssignment(object):
    _Y2KReference = datetime.datetime(2000, 1, 1)

    def __init__(self, reading_id, utc, found, exact, crossed_break):
        self.reading_id = reading_id
        self.utc = utc
        self.found_id = found
        self.exact = exact
        self.crossed_break = crossed_break

        rtc_value = int((utc - self._Y2KReference).total_seconds())
        self.rtc_value = rtc_value | (1 << 31)

    def __str__(self):
        return "%s (reading_id=%08X, exact=%s, crossed_break=%s)" % (self.utc, self.reading_id, self.exact, self.crossed_break)


class UTCAssigner(object):
    """Helper class to assign UTC timestamps to device readings."""

    _Y2KReference = datetime.datetime(2000, 1, 1)

    def __init__(self):
        self._anchor_points = SortedKeyList(key=lambda x: x.reading_id)
        self._anchor_streams = {}
        self._break_streams = set()

        self._known_converters = {
            'rtc': UTCAssigner.convert_rtc
        }

    def _load_known_breaks(self):
        self._break_streams.add(0x5C00)

    def anchor_stream(self, stream_id, converter="rtc"):
        """Mark a stream as containing anchor points."""

        if isinstance(converter, str):
            converter = self._known_converters.get(converter)

            if converter is None:
                raise ArgumentError("Unknown anchor converter string: %s" % converter, known_converters=list(self._known_converters))

        self._anchor_streams[stream_id] = converter

    @classmethod
    def convert_rtc(cls, timestamp):
        """Convert a number of seconds since 1/1/2000 to UTC time."""

        if timestamp & (1  << 31):
            timestamp &= ~(1 << 31)

        delta = datetime.timedelta(seconds=timestamp)
        return cls._Y2KReference + delta

    def add_point(self, reading_id, uptime=None, utc=None, is_break=False):
        """Add a time point that could be used as a UTC reference."""

        if reading_id == 0:
            return

        if uptime is None and utc is None:
            return

        if uptime & (1 << 31):
            if utc is not None:
                return

            uptime &= ~(1 << 31)

            utc = self.convert_rtc(uptime)
            uptime = None

        anchor = _TimeAnchor(reading_id, uptime, utc, is_break)

        if anchor in self._anchor_points:
            return

        self._anchor_points.add(anchor)

    def add_reading(self, reading):
        """Add an IOTileReading."""

        is_break = False
        utc = None

        if reading.stream in self._break_streams:
            is_break = True

        if reading.stream in self._anchor_streams:
            utc = self._anchor_streams[reading.stream](reading)

        self.add_point(reading.reading_id, reading.raw_time, utc, is_break=is_break)

    def add_report(self, report, ignore_errors=False):
        """Add all anchors from a report."""

        if not isinstance(report, SignedListReport):
            if ignore_errors:
                return

            raise ArgumentError("You can only add SignedListReports to a UTCAssigner", report=report)

        for reading in report.visible_readings:
            self.add_reading(reading)

        self.add_point(report.report_id, report.sent_timestamp, report.received_time)

    def assign_utc(self, reading_id, uptime=None):
        """Assign a utc datetime to a reading id.

        This method will return an object with assignment information or None
        if a utc value cannot be assigned.  The assignment object returned
        contains a utc property that has the asssigned UTC as well as other
        properties describing how reliable the assignment is.
        """

        if len(self._anchor_points) == 0:
            return None

        if reading_id > self._anchor_points[-1].reading_id:
            return None

        i = self._anchor_points.bisect_key_left(reading_id)
        found_id = False
        crossed_break = False
        exact = True

        last = self._anchor_points[i].copy()
        if uptime is not None:
            last.uptime = uptime

        if last.reading_id == reading_id:
            found_id = True

            if last.utc is not None:
                return last.utc

        accum_delta = 0
        for curr in self._anchor_points.islice(i + 1):
            if curr.uptime is None or last.uptime is None:
                exact = False
            elif curr.is_break or curr.uptime < last.uptime:
                exact = False
                crossed_break = True
            else:
                accum_delta += curr.uptime - last.uptime

            if curr.utc is not None:
                time_delta = datetime.timedelta(seconds=accum_delta)
                return UTCAssignment(reading_id, curr.utc - time_delta, found_id, exact, crossed_break)

            last = curr

        return None



from .app import IOTileApp
from .proxy import TileBusProxyObject
from .hwmanager import HardwareManager
from .update import UpdateScript

__all__ = ['IOTileApp', 'TileBusProxyObject', 'HardwareManager']



from typedargs.exceptions import ArgumentError
from .proxy import TileBusProxyObject

class TileBusProxyPlugin(object):
    """
    Proxy plugin objects are mixin style classes that add functionality modules to proxy objects
    """

    def __init__(self, parent):
        if not isinstance(parent, TileBusProxyObject):
            raise ArgumentError("Attempting to initialize a TileBusProxyPlugin with an invalid parent object", parent=parent)

        self._proxy = parent

    def rpc(self, feature, cmd, *args, **kw):
        return self._proxy.rpc(feature, cmd, *args, **kw)



"""All currently support record types for remote update scripts."""

from .reflash_tile import ReflashTileRecord
from .unknown import UnknownRecord
from .send_rpc import SendRPCRecord, SendErrorCheckingRPCRecord
from .reflash_controller import ReflashControllerRecord
from .reset_device import ResetDeviceRecord
from .set_version import SetDeviceTagRecord
from ..record import UpdateRecord

# Register all of our record types to be decodable (we don't register UnknownRecord)
UpdateRecord.RegisterRecordType(ReflashTileRecord)
UpdateRecord.RegisterRecordType(SendRPCRecord)
UpdateRecord.RegisterRecordType(SendErrorCheckingRPCRecord)
UpdateRecord.RegisterRecordType(ReflashControllerRecord)
UpdateRecord.RegisterRecordType(ResetDeviceRecord)
UpdateRecord.RegisterRecordType(SetDeviceTagRecord)

__all__ = ['ReflashTileRecord', 'UnknownRecord', 'SendRPCRecord', 'SendErrorCheckingRPCRecord', 'ReflashControllerRecord', 'SetDeviceTagRecord']



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

#exceptions.py

import iotile.core.exceptions


class RPCError(iotile.core.exceptions.HardwareError):
    pass

class ModuleBusyError(RPCError):
    def __init__(self, address, **kwargs):
        super(ModuleBusyError, self).__init__("Module responded that it was busy", address=address, **kwargs)

class UnsupportedCommandError(RPCError):
    def __init__(self, **kwargs):
        super(UnsupportedCommandError, self).__init__("Module did not support the specified command", **kwargs)

class ModuleNotFoundError(RPCError):
    def __init__(self, address, **kwargs):
        super(ModuleNotFoundError, self).__init__("Module was not found or did not respond", address=address, **kwargs)

class StreamNotConnectedError(RPCError):
    def __init__(self, **kwargs):
        super(StreamNotConnectedError, self).__init__("Stream was not connected to any MoMo devices", **kwargs)

class StreamOperationNotSupportedError(RPCError):
    def __init__(self, **kwargs):
        super(StreamOperationNotSupportedError, self).__init__("Underlying command stream does not support the required operation", **kwargs)

class InvalidReturnValueError(RPCError):
    """Exception thrown when the return value of an RPC does not conform to its known format."""

    def __init__(self, address, rpc_id, format_code, response, **kwargs):
        super(InvalidReturnValueError, self).__init__("Invalid return value from Tile %d, RPC 0x%04x, expected format %s, got %d bytes"
                                                      % (address, rpc_id, format_code, len(response)), **kwargs)

        self.address = address
        self.rpc_id = rpc_id
        self.format_code = format_code
        self.response = response

class UnknownModuleTypeError(iotile.core.exceptions.TypeSystemError):
    pass



"""Base class for all IOtile Apps."""


class IOTileApp(object):
    """Base class for all IOTile Apps.

    An IOTile App is a class that represents all of the externally visible
    functionality of an IOTile device.  It is typically a subset of what is
    exposed on the individual tiles that make up the iotile device with a
    higher level API.  For example, there may be a shipping tracker IOTile
    device that has an accelerometer and an environmental tile.  The proxy
    objects for those tiles will expose a lot of functionality related to
    acceleration and environmental readings but nothing related to tracking
    a shipment.

    An IOTile App ties together the behavior of all of its underlying tiles.

    IOTile Apps are matched with devices using the device's app tag and version
    which is a number that unique identifies the app running on the device and
    a major.minor version number.

    All IOTile Apps are required to implement a MatchInfo module function if
    they support automatic matching with a device.  The MatchInfo function
    should return a list of (tag, SemanticVersionRange, quality) tuples for each
    app tag they should match with.  The quality value is a floating point
    number between 0 and 100, inclusive, where higher quality matches take
    priority over lower quality matches.  A good default choice is 50.

    Args:
        hw (HardwareManager): A HardwareManager instance connected to a
            matching device.
        app_info (tuple): The app_tag and version of the device we are
            connected to.
        os_info (tuple): The os_tag and version of the device we are
            connected to.
        device_id (int): The UUID of the device that we are connected to.
    """

    def __init__(self, hw, app_info, os_info, device_id):
        self._hw = hw
        self._app_tag = app_info[0]
        self._app_version = app_info[1]
        self._os_tag = os_info[0]
        self._os_version = os_info[1]
        self._device_id = device_id

    @classmethod
    def MatchInfo(cls):
        """Return a list of matching app tag and versions.

        This method must be overriden by all subclasses to return
        their matching information if they want to support automatically
        matching with a device by app_tag.

        Returns:
            list of (int, SemanticVersionRange, float) tuples: A tuple with
                the app_tag and matching version range for all app tags that
                this IOTileApp should match with.  The third float is a
                quality number that will be used to break ties when more than
                one app matches a given device.  Higher is better and the
                value should be in the range [0, 100].
        """

        return []

    @classmethod
    def AppName(cls):
        """A unqiue name for this app so that it can be loaded by name.

        Returns:
            str: The unique name for this app module.
        """

        raise NotImplementedError()



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

#MIB Proxy Objects

from builtins import range
from iotile.core.hw.exceptions import *
from iotile.core.utilities.typedargs import return_type, annotated, param, context
from time import sleep
import struct
from iotile.core.exceptions import *
from ..virtual import unpack_rpc_payload
from builtins import str, int


class TileBusProxyObject(object):
    def __init__(self, stream, address):
        self.stream = stream
        self.addr = address
        self._config_manager = ConfigManager(self)
        self._hwmanager = None

    @classmethod
    def ModuleName(cls):
        return 'NO APP'

    @annotated
    def config_manager(self):
        return self._config_manager

    def rpc(self, feature, cmd, *args, **kw):
        """
        Send an RPC call to this module, interpret the return value
        according to the result_type kw argument.  Unless raise keyword
        is passed with value False, raise an RPCException if the command
        is not successful.
        """

        rpc_id = (feature << 8 | cmd)

        if 'arg_format' in kw:
            packed_args = struct.pack("<{}".format(kw['arg_format']), *args)
        elif not args:
            packed_args = b''
        else:
            packed_args = self._format_args(args)

        status, payload = self.stream.send_rpc(self.addr, rpc_id, packed_args, **kw)

        unpack_flag = False
        if "result_type" in kw:
            res_type = kw['result_type']
        elif "result_format" in kw:
            unpack_flag = True
            res_type = (0, True)
        else:
            res_type = (0, False)

        try:
            res = self._parse_rpc_result(status, payload, *res_type, command=rpc_id)
            if unpack_flag:
                try:
                    return unpack_rpc_payload("%s" % kw["result_format"], res['buffer'])
                except struct.error as err:
                    raise InvalidReturnValueError(self.addr, rpc_id, kw['result_format'], res['buffer'], status=status)

            return res
        except ModuleBusyError:
            pass

        if "retries" not in kw:
            kw['retries'] = 10

        #Sleep 100 ms and try again unless we've exhausted our retry attempts
        if kw["retries"] > 0:
            kw['retries'] -= 1

            sleep(0.1)
            return self.rpc(feature, cmd, *args, **kw)

    def _convert_int(self, arg):
        out = bytearray(2)

        out[0] = arg & 0xFF
        out[1] = (arg & 0xFF00) >> 8

        converted = out[0] | (out[1] << 8)

        if converted != arg:
            raise ValueError("Integer argument was too large to fit in an rpc 16 bit int: %d" % arg)

        return out

    def _pack_arg(self, arg):
        if isinstance(arg, int):
            return self._convert_int(arg), False
        elif isinstance(arg, bytearray):
            return arg, True
        elif isinstance(arg, str):  # for python 3 compatibility, encode all newstr from future module
            return bytearray(arg.encode('utf-8')), True
        elif isinstance(arg, bytes):
            return bytearray(arg), True

        raise ValueError("Unknown argument type could not be converted for rpc call.")

    def _format_args(self, args):
        fmtd = bytearray()

        num_ints = 0
        num_bufs = 0

        for arg in args:
            a, is_buf = self._pack_arg(arg)
            fmtd += a

            if is_buf:
                num_bufs += 1
                buff_len = len(a)

            if not is_buf:
                if num_bufs != 0:
                    raise ValueError("Invalid rpc parameters, integer came after buffer.")

                num_ints += 1

        if num_bufs > 1:
            raise ValueError("You must pass at most 1 buffer. num_bufs=%d" % num_bufs)

        if len(fmtd) > 20:
            raise ValueError("Arguments are greater then the maximum mib packet size, size was %d" % len(fmtd))

        return fmtd

    def rpc_v2(self, cmd, arg_format, result_format, *args, **kw):
        """
        Send an RPC call to this module, interpret the return value
        according to the result_type kw argument.  Unless raise keyword
        is passed with value False, raise an RPCException if the command
        is not successful.

        v2 enforces the use of arg_format and result_format
        v2 combines the feature+cmd chunks in to a single 4-byte chunk
        """
        packed_args = struct.pack("<{}".format(arg_format), *args)
        status, payload = self.stream.send_rpc(self.addr, cmd, packed_args, **kw)
        res_type = (0, True)

        try:
            res = self._parse_rpc_result(status, payload, *res_type, command=cmd)
            return unpack_rpc_payload("%s" % result_format, res['buffer'])
        except ModuleBusyError:
            pass

        if "retries" not in kw:
            kw['retries'] = 10

        #Sleep 100 ms and try again unless we've exhausted our retry attempts
        if kw["retries"] > 0:
            kw['retries'] -= 1

            sleep(0.1)
            return self.rpc_v2(cmd, arg_format, result_format, *args, **kw)


    @return_type("string")
    def hardware_version(self):
        """Return the embedded hardware version string for this tile.

        The hardware version is an up to 10 byte user readable string that is
        meant to encode any necessary information about the specific hardware
        that this tile is running on.  For example, if you have multiple
        assembly variants of a given tile, you could encode that information
        here.

        Returns:
            str: The hardware vesrion read from the tile.
        """
        res = self.rpc(0x00, 0x02, result_type=(0, True))

        #Result is a string but with zero appended to the end to make it a fixed 10 byte
        #size
        binary_version = res['buffer']

        ver = ""

        for x in binary_version:
            if x != 0:
                ver += chr(x)

        return ver

    @param("expected", "string", desc="The hardware string we expect to find")
    @return_type("bool")
    def check_hardware(self, expected):
        """Make sure the hardware version is what we expect.

        This convenience function is meant for ensuring that we are talking to
        a tile that has the correct hardware version.

        Args:
            expected (str): The expected hardware string that is compared
                against what is reported by the hardware_version RPC.

        Returns:
            bool: true if the hardware is the expected version, false otherwise
        """

        if len(expected) < 10:
            expected += '\0'*(10 - len(expected))

        err, = self.rpc(0x00, 0x03, expected, result_format="L")
        if err == 0:
            return True

        return False

    @return_type("basic_dict")
    def status(self):
        """
        Query the status of an IOTile including its name and version
        """

        hw_type, name, major, minor, patch, status = self.rpc(0x00, 0x04, result_format="H6sBBBB")

        status = {
            'hw_type': hw_type,
            'name': name.decode('utf-8'),
            'version': (major, minor, patch),
            'status': status
        }

        return status

    @param("wait", "float", desc="Time to wait after reset for tile to boot up to a usable state")
    def reset(self, wait=1.0):
        """
        Immediately reset this tile.
        """
        try:
            self.rpc(0x00, 0x01)
        except ModuleNotFoundError:
            pass

        sleep(wait)

    @return_type("string")
    def tile_name(self):
        stat = self.status()

        return stat['name']

    @return_type("list(integer)", formatter="compact")
    def tile_version(self):
        stat = self.status()

        return stat['version']

    @return_type("map(string, bool)")
    def tile_status(self):
        """
        Get the current status of this tile

        Returns a
        """
        stat = self.status()

        flags = stat['status']

        #FIXME: This needs to stay in sync with lib_common: cdb_status.h
        status = {}
        status['debug_mode'] = bool(flags & (1 << 3))
        status['configured'] = bool(flags & (1 << 1))
        status['app_running'] = bool(flags & (1 << 0))
        status['trapped'] = bool(flags & (1 << 2))

        return status


    def _parse_rpc_result(self, status, payload, num_ints, buff, command):
        """
        Parse the response of an RPC call into a dictionary with integer and buffer results
        """

        parsed = {'ints':[], 'buffer':"", 'error': 'No Error', 'is_error': False}
        parsed['status'] = status
        parsed['return_value'] = status & 0b00111111

        # Check for protocol defined errors
        if not status & (1<<6):
            if status == 2:
                raise UnsupportedCommandError(address=self.addr, command=command)

            raise RPCError("Unknown status code received from RPC call", address=self.addr, status_code=status)

        # This second check below is a workaround for a lib_controller bug introduced
        # 9/27/2017 that causes the app_defined status bit to be set for all status
        # codes that come from a controller.
        if self.addr == 8 and status == ((1 << 6) | 2):
            raise UnsupportedCommandError(address=self.addr, command=command)

        #Otherwise, parse the results according to the type information given
        size = len(payload)

        if size < 2*num_ints:
            raise RPCError('Return value too short to unpack', expected_minimum_size=2*num_ints, actual_size=size, status_code=status, payload=payload)
        elif buff is False and size != 2*num_ints:
            raise RPCError('Return value was not the correct size', expected_size=2*num_ints, actual_size=size, status_code=status, payload=payload)

        for i in range(0, num_ints):
            low = (payload[2*i])
            high = (payload[2*i + 1])
            parsed['ints'].append((high << 8) | low)

        if buff:
            parsed['buffer'] = payload[2*num_ints:]

        return parsed


@context("ConfigManager")
class ConfigManager(object):
    """
    Manager Proxy for configuration variables on IOTiles

    Handles querying what config variables are defined, setting and getting
    their values.  Note that config variables should not change when application
    code is running so set_config_variables should not be called once an application
    is launched.

    Config variables should be thought of like environment variables in UNIX.  Each
    tile is conceptually a process and the config variables are the environment
    variables that are used to configure how that process runs.  They are captured
    when the process starts and should not change during process execution.

    Similarly, config variables are set on each tile by the TileBus controller every
    time the tile resets.  The tile can then read them to figure out how it should
    initialize itself.

    The variable cannot change value while the tile is running.  If you want to update
    a config variable, you need to stream that new value to the tile immediately after
    it resets before it passes control from its executive to its application firmware.

    This is usually not done manually but rather as part of the initialization process
    performed by the TileBus controller managing the IOTile Device.

    This ConfigManager is useful primarily for debugging, checking what values are actually
    set on a tile and what varibles it supports.
    """

    def __init__(self, parent):
        self._proxy = parent

    @return_type('list(integer)')
    def list_variables(self):
        """ List all of the configuration variables defined by this tile

        This function will tell you every configuration variable that this tile
        defines.  You can then query more information about a given variable
        using the returned 16 bit ids with the functions `describe_variable` and
        `get_variable`.

        Returns:
            list: A list of the ids of all of the config variables that are
            supported by this tile.
        """

        offset = 0
        ids = []

        while True:
            resp = self._proxy.rpc(0, 10, offset, result_type=(1, True))
            count = resp['ints'][0]

            if count == 0:
                break

            fmt = "<%dH" % count
            id_chuck = struct.unpack(fmt, resp['buffer'][:2*count])

            ids += id_chuck

            if count != 9:
                break

            offset += 9

        return ids

    @return_type("basic_dict")
    @param("id", "integer", desc="Variable ID to describe")
    def describe_variable(self, id):
        """Describe a configuration variable

        Queries metadata about a config variable directly from the tile where it
        is defined.  This information includes its size, where it is stored in RAM,
        and whether it has a fixed or variable length.

        The config variable is identified by its 16-bit numeric id that should match
        the value used to define the variable during the firmware build process in
        a .bus or .cdb file.

        Params:
            id (int): The 16 bit id of the config variable that we are trying to
                get information about.

        Returns:
            dict: A dict with information describing the config variable including
                its address in RAM, 16-bit id, maximum size and whether the variable
                is a fixed or variable size.
        """

        resp = self._proxy.rpc(0, 11, id, result_type=(2, True))

        err = resp['ints'][0]
        if err != 0:
            raise HardwareError("Error finding config variable by id", id=id, error_code=err)

        addr, id, flags = struct.unpack("<LHH", resp['buffer'])

        maxsize = (flags & ~(1 << 15)) & 0xFFFF
        variable_size = bool(flags >> 15)

        info = {
            'address': addr,
            'id': id,
            'max_size': maxsize,
            'variable_size': variable_size
        }

        return info

    @return_type("bytes", "repr")
    @param("id", "integer", desc="Variable ID to fetch")
    def get_variable(self, id):
        """Get value stored in a config variable

        Params:
            id (int): The 16 bit id of the config variable that we are trying to
                get the value of.

        Returns:
            bytearray: A raw byte array containing the binary data contained
                in this config variable.
        """

        offset = 0
        resp = self._proxy.rpc(0, 13, id, offset, result_type=(0, True))
        if len(resp['buffer']) == 0:
            return bytearray()

        retval = resp['buffer']

        while len(resp['buffer']) > 0:
            offset = len(retval)
            resp = self._proxy.rpc(0, 13, id, offset, result_type=(0, True))

            retval += resp['buffer']

        return retval

    @param("id", "integer", desc="Variable ID to set")
    @param("value", "bytes", desc="hexadecimal byte value to set")
    def set_variable(self, id, value):
        """Set the value stored in a config variable

        This function wraps the underlying RPC that is used internally in
        IOTile devices for the controller to update a tile's config variables
        every time the device resets.

        Since config variables are forbidden from changing while a tile is running,
        using this function directly is discouraged since it's unlikely you will
        calling it before the tile excecutive passes controller to its application.

        It is included here mainly for completeness and for advanced use cases
        where you need to directly set a config variable.

        Normally you should set config variables by adding an entry to the config
        database on the controller, which will then stream the data to the correct
        tile every time the tile resets and registers itself with the controller.

        Params:
            id (int): The 16 bit id of the config variable that we are trying to
                get the value of.
            value (bytearray): The data to store in the config variable
        """

        for offset in range(0, len(value), 16):
            remaining = len(value) - offset
            if remaining > 16:
                remaining = 16

            resp = self._proxy.rpc(0, 12, id, offset, value[offset:offset+remaining], result_type=(1, False))
            if resp['ints'][0] != 0:
                raise HardwareError("Error setting config variable", id=id, error_code=resp['ints'][0])



"""A Basic authentication provider that can just sign with SHA256 hash."""

import hashlib
import hmac
from iotile.core.exceptions import NotFoundError
from .auth_provider import AuthProvider


class BasicAuthProvider(AuthProvider):
    """Basic default authentication provider that can only calculate sha256 hashes
    """

    def sign_report(self, device_id, root, data, **kwargs):
        """Sign a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should sign
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signature itself must always be a bytearray stored under the
                'signature' key, however additional keys may be present depending
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to sign the data.
        """

        AuthProvider.VerifyRoot(root)

        if root != AuthProvider.NoKey:
            raise NotFoundError('unsupported root key in BasicAuthProvider', root_key=root)

        result = bytearray(hashlib.sha256(data).digest())
        return {'signature': result, 'root_key': root}

    def verify_report(self, device_id, root, data, signature, **kwargs):
        """Verify a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should verify
            signature (bytearray): The signature attached to data that we should verify
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to verify the data due to
                an error.  If the data is simply not valid, then the function returns
                normally.
        """

        AuthProvider.VerifyRoot(root)

        if root != AuthProvider.NoKey:
            raise NotFoundError('unsupported root key in BasicAuthProvider', root_key=root)


        result = bytearray(hashlib.sha256(data).digest())

        if len(signature) == 0:
            verified = False
        elif len(signature) > len(result):
            verified = False
        elif len(signature) < len(result):
            trunc_result = result[:len(signature)]
            verified = hmac.compare_digest(signature, trunc_result)
        else:
            verified = hmac.compare_digest(signature, result)

        return {'verified': verified, 'bit_length': 8*len(signature)}



from iotile.core.exceptions import *

# Generate a 0000-0000-0000-0001 given an integer
int64gid = lambda n: '-'.join(['{:04x}'.format(n >> (i << 4) & 0xFFFF) for i in range(0, 4)[::-1]])


def uuid_to_slug(id):
    """
    Return IOTile Cloud compatible Device Slug

    :param id: UUID
    :return: string in the form of d--0000-0000-0000-0001
    """
    if not isinstance(id, int):
        raise ArgumentError("Invalid id that is not an integer", id=id)

    if id < 0 or id > 0x7fffffff:
        # For now, limiting support to a signed integer (which on some platforms, can be 32bits)
        raise ArgumentError("Integer should be a positive number and smaller than 0x7fffffff", id=id)

    return '--'.join(['d', int64gid(id)])



"""A thread that hands workqueue items to a user function."""

import threading
import logging
import sys
from collections import namedtuple
from queue import Queue
from future.utils import raise_
from iotile.core.exceptions import TimeoutExpiredError

STOP_WORKER_ITEM = object()
MarkLocationItem = namedtuple('MarkLocationItem', ['callback'])
WorkItem = namedtuple("WorkItem", ['arg', 'callback'])
WaitIdleItem = namedtuple("WaitIdleItem", ['callback'])


class WorkQueueThread(threading.Thread):
    """A worker thread that handles queued work.

    This class takes a single callable as a parameter in __init__
    and starts a background thread with a shared queue.  Whenever
    an item is put in the shared queue, the callable is called with
    that work item.

    The return value of that callable is passed back to the caller
    or given via a callback function.

    The key methods of this class are:

    - dispatch(item, callback=None): Synchronously dispatches a work item to
      the queue and waits for the result.  If callback is not None, this
      function will return immediately and the result will be passed to the
      callback when it is ready.
    - stop(): Synchronously stop the background thread.  This will wait for
      all currently queued work items to finish and then cleanly shut down the
      background thread.
    - flush(): Wait until the work queue is momentarily empty().  This method
      is useful for ensuring that all work items received up until this method
      was called have been processed.

    Args:
        handler (callable): The handler function that will be pasesd all of the
            work items queued in dispatch() and should return a result that will
            be the return value of dispatch().  If this function throws an exception,
            it will be rethrown from dispatch.
    """

    def __init__(self, handler):
        super(WorkQueueThread, self).__init__()
        self.daemon = True

        self._routine = handler
        self._work_queue = Queue()
        self._logger = logging.getLogger(__name__)

    def dispatch(self, value, callback=None):
        """Dispatch an item to the workqueue and optionally wait.

        This is the only way to add work to the background work queue. Unless
        you also pass a callback object, this method will synchronously wait
        for the work to finish and return the result. If the work raises an
        exception, the exception will be reraised in this method.

        If you pass an optional callback(exc_info, return_value), this method
        will not block and instead your callback will be called when the work
        finishes.  If an exception was raised during processing, exc_info will
        be set with the contents of sys.exc_info().  Otherwise, exc_info will
        be None and whatever the work_queue handler returned will be passed as
        the return_value parameter to the supplied callback.

        Args:
            value (object): Arbitrary object that will be passed to the work
                queue handler.

            callback (callable): Optional callback to receive the result of
                the work queue when it finishes.  If not passed, this method
                will be synchronous and return the result from the dispatch()
                method itself

        Returns:
            object: The result of the work_queue handler function or None.

            If callback is not None, then this method will return immediately
            with a None return value.  Otherwise it will block until the work
            item is finished (including any work items ahead in the queue) and
            return whatever the work item handler returned.
        """

        done = None

        if callback is None:
            done = threading.Event()
            shared_data = [None, None]

            def _callback(exc_info, return_value):
                shared_data[0] = exc_info
                shared_data[1] = return_value

                done.set()

            callback = _callback

        workitem = WorkItem(value, callback)
        self._work_queue.put(workitem)
        if done is None:
            return None

        done.wait()
        exc_info, return_value = shared_data
        if exc_info is not None:
            raise_(*exc_info)  #pylint:disable=not-an-iterable;We know it is iterable when done is not None

        return return_value

    def flush(self):
        """Synchronously wait until this work item is processed.

        This has the effect of waiting until all work items queued before this
        method has been called have finished.
        """

        done = threading.Event()

        def _callback():
            done.set()

        self.defer(_callback)
        done.wait()

    def defer(self, callback):
        """Schedule a callback once all current items in the queue are finished.

        This function can schedule work synchronous with the work queue without
        passing through the work queue handler.

        This method returns immediately.

        Args:
            callback (callable): A callable with no arguments that will be
                called once all current items in the workqueue have been
                executed.
        """

        self._work_queue.put(MarkLocationItem(callback))

    def defer_until_idle(self, callback):
        """Wait until the work queue is (temporarily) empty.

        This is different from flush() because processing a work queue entry
        may add additional work queue entries.  This method lets you wait
        until there are no more entries in the work queue.

        Depending on how work is being added to the work queue, this may be
        a very interesting condtion.

        This method will return immeidately and schedule callback to be called
        as soon as the work queue becomes empty.  You can queue as many
        callbacks as you like via multiple calls to defer_until_idle. These
        will be executed in the same order together the first time that the
        queue becomes momentarily idle.

        Note that the concept of an "empty" workqueue is a very unstable
        concept in general.  Unless you as the caller know that no one else
        except you and possibly the work-queue items themselves can add a task
        to the work queue, then there is no guarantee that this callback will
        ever fire since it could be that someone else is adding work queue
        items just as fast as they are being completed.

        This is a specialty method that is useful in a few defined
        circumstances.

        Args:
            callback (callable): A callable with no arguments that will be
                called once the queue is temporarily empty.
        """

        self._work_queue.put(WaitIdleItem(callback))

    def wait_until_idle(self):
        """Block the calling thread until the work queue is (temporarily) empty.

        See the detailed discussion under defer_until_idle() for restrictions
        and expected use cases for this method.

        This routine will block the calling thread.
        """

        done = threading.Event()

        def _callback():
            done.set()

        self.defer_until_idle(_callback)
        done.wait()

    def run(self):
        """The target routine called to start thread activity."""

        idle_watchers = []

        while True:
            try:
                if self._work_queue.empty() and len(idle_watchers) > 0:
                    for watcher in idle_watchers:
                        try:
                            watcher()
                        except: #pylint:disable=bare-except;We can't let one idle watcher failure impact any other watcher
                            self._logger.exception("Error inside queue idle watcher")

                    idle_watchers = []

                item = self._work_queue.get()

                # Handle special actions that are not RPCs
                if item is STOP_WORKER_ITEM:
                    return
                elif isinstance(item, MarkLocationItem):
                    item.callback()
                    continue
                elif isinstance(item, WaitIdleItem):
                    idle_watchers.append(item.callback)
                    continue
                elif not isinstance(item, WorkItem):
                    self._logger.error("Invalid item passed to WorkQueueThread: %s, ignoring", item)
                    continue

                try:
                    exc_info = None
                    retval = None

                    retval = self._routine(item.arg)
                except:  #pylint:disable=bare-except;We need to capture the exception and feed it back to the caller
                    exc_info = sys.exc_info()

                if item.callback is not None:
                    item.callback(exc_info, retval)
            except:  #pylint:disable=bare-except;We cannot let this background thread die until we are told to stop()
                self._logger.exception("Error inside background workqueue thread")

    def stop(self, timeout=None, force=False):
        """Stop the worker thread and synchronously wait for it to finish.

        Args:
            timeout (float): The maximum time to wait for the thread to stop
                before raising a TimeoutExpiredError.  If force is True, TimeoutExpiredError
                is not raised and the thread is just marked as a daemon thread
                so that it does not block cleanly exiting the process.
            force (bool): If true and the thread does not exit in timeout seconds
                no error is raised since the thread is marked as daemon and will
                be killed when the process exits.
        """

        self.signal_stop()
        self.wait_stopped(timeout, force)

    def signal_stop(self):
        """Signal that the worker thread should stop but don't wait.

        This function is useful for stopping multiple threads in parallel when
        combined with wait_stopped().
        """

        self._work_queue.put(STOP_WORKER_ITEM)

    def wait_stopped(self, timeout=None, force=False):
        """Wait for the thread to stop.

        You must have previously called signal_stop or this function will
        hang.

        Args:

            timeout (float): The maximum time to wait for the thread to stop
                before raising a TimeoutExpiredError.  If force is True,
                TimeoutExpiredError is not raised and the thread is just
                marked as a daemon thread so that it does not block cleanly
                exiting the process.
            force (bool): If true and the thread does not exit in timeout seconds
                no error is raised since the thread is marked as daemon and will
                be killed when the process exits.
        """

        self.join(timeout)

        if self.is_alive() and force is False:
            raise TimeoutExpiredError("Error waiting for background thread to exit", timeout=timeout)



from __future__ import unicode_literals, absolute_import, print_function
import struct
from future.utils import python_2_unicode_compatible
from iotile.core.exceptions import ArgumentError
from ..record import MatchQuality
from .send_rpc import SendErrorCheckingRPCRecord


@python_2_unicode_compatible
class SetDeviceTagRecord(SendErrorCheckingRPCRecord):
    """Set the device's app or os tag and version.

    Args:
        app_tag (int): The app tag we wish to set
        app_version (str): The X.Y version that we wish to set for the app
        os_tag (int): The os tag we wish to set
        os_version (str): The X.Y os version we wish to set
    """

    RPC_ID = 0x100B

    def __init__(self, app_tag=None, app_version=None, os_tag=None, os_version=None):
        update_app = app_tag is not None
        update_os = os_tag is not None

        if app_version is None:
            app_version = "0.0"
        if os_version is None:
            os_version = "0.0"

        app_info = _combine_info(app_tag, app_version)
        os_info = _combine_info(os_tag, os_version)

        payload = struct.pack("<LLBB", os_info, app_info, int(update_os), int(update_app))

        self.update_app = update_app
        self.update_os = update_os
        self.app_tag = app_tag
        self.app_version = app_version
        self.os_tag = os_tag
        self.os_version = os_version

        super(SetDeviceTagRecord, self).__init__(8, SetDeviceTagRecord.RPC_ID, payload, response_size=4)

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this functon should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): The number of binary records that are included
                in record_data.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.
        """

        if record_count > 1:
            return MatchQuality.NoMatch

        cmd, _address, _resp_length, payload = cls._parse_rpc_info(record_data)

        if cmd == cls.RPC_ID:
            try:
                _os_info, _app_info, update_os, update_app = struct.unpack("<LLBB", payload)
                update_os = bool(update_os)
                update_app = bool(update_app)

            except ValueError:
                return MatchQuality.NoMatch

            return MatchQuality.PerfectMatch

        return MatchQuality.NoMatch

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This should be called with a binary record blob (NOT including the
        record type header) and it will decode it into a SetDeviceTagRecord.

        Args:
            record_data (bytearray): The raw record data that we wish to parse
                into an UpdateRecord subclass NOT including its 8 byte record header.
            record_count (int): The number of records included in record_data.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.

        Returns:
            SetDeviceTagRecord: The decoded reflash tile record.
        """

        _cmd, _address, _resp_length, payload = cls._parse_rpc_info(record_data)

        try:
            os_info, app_info, update_os, update_app = struct.unpack("<LLBB", payload)
            update_os = bool(update_os)
            update_app = bool(update_app)

            if update_app and not update_os:
                tag, version = _parse_info(app_info)
                return SetDeviceTagRecord(app_tag=tag, app_version=version)
            elif update_os and not update_app:
                tag, version = _parse_info(os_info)
                return SetDeviceTagRecord(os_tag=tag, os_version=version)
            elif update_os and update_app:
                os_tag, os_version = _parse_info(os_info)
                app_tag, app_version = _parse_info(app_info)
                return SetDeviceTagRecord(app_tag=app_tag, app_version=app_version, os_tag=os_tag, os_version=os_version)
            else:
                raise ArgumentError("Neither update_os nor update_app is set True")

        except ValueError:
            raise ArgumentError("Could not parse set device version payload", payload=payload)

    def __str__(self):
        if self.update_app and not self.update_os:
            return "Set device app to (tag:%d version:%s)" % (self.app_tag, self.app_version)
        elif self.update_os and not self.update_app:
            return "Set device os to (tag:%d, version:%s)" % (self.os_tag, self.os_version)
        elif self.update_os and self.update_app:
            return "Set device os to (tag:%d, version:%s) and app to (tag:%d, version:%s)" \
                % (self.os_tag, self.os_version, self.app_tag, self.app_version)
        else:
            raise ArgumentError("Neither update_os nor update_app is set True")


def _parse_info(combined_info):
    tag = combined_info & ((1 << 20) - 1)

    major = combined_info >> 26 & ((1 << 6) - 1)
    minor = combined_info >> 20 & ((1 << 6) - 1)

    return tag, "%d.%d" % (major, minor)


def _combine_info(tag, version):
    if version is None:
        version = "0.0"

    if tag is None:
        tag = 0

    major, _, minor = version.partition(".")

    if tag >= (1 << 20):
        raise ArgumentError("The tag number is too high.  It must fit in 20-bits", max_tag=1 << 20, tag=tag)

    if "." not in version:
        raise ArgumentError("You must pass a version number in X.Y format", version=version)

    major, _, minor = version.partition('.')
    try:
        major = int(major)
        minor = int(minor)
    except ValueError:
        raise ArgumentError("Unable to convert version string into major and minor version numbers", version=version)

    if major < 0 or minor < 0 or major >= (1 << 6) or minor >= (1 << 6):
        raise ArgumentError("Invalid version numbers that must be in the range [0, 63]", major=major, minor=minor, version_string=version)

    version_number = (major << 6) | minor
    combined_tag = (version_number << 20) | tag

    return combined_tag



"""A command line script to print the contents of an UpdateScript."""

from __future__ import unicode_literals, absolute_import, print_function
import sys
import argparse
import logging
from iotile.core.hw import UpdateScript
from iotile.core.exceptions import ArgumentError, DataError

DESCRIPTION = \
"""Print information about an IOTile update script.

This program takes in a binary UpdateScript object, parses it and prints out
information about it.  It is able to decode the script into the series of high
level operations that it contains including decoding multi-step operations
back into their high level logical equivalent such as 'Program SensorGraph'
rather than a series of 'Add Node' and 'Add Streamer' actions.
"""

def build_args():
    """Create command line parser."""

    parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('script', type=str, help="The binary update script to load")
    parser.add_argument('-v', '--verbose', action="count", help="Increase logging level (goes error, warn, info, debug)")
    parser.add_argument('--show-rpcs', '-r', action="store_true", help="Show TRUB steps as RPC calls where applicable")
    parser.add_argument('--allow-unknown', '-u', action="store_true", help="Don't complain if the script contains unknown actions")
    parser.add_argument('-f', '--format', choices=['text', 'json'], default="text", help="The output format to use to display the script")

    return parser


def main(argv=None):
    """Main script entry point.

    Args:
        argv (list): The command line arguments, defaults to sys.argv if not passed.

    Returns:
        int: The return value of the script.
    """

    if argv is None:
        argv = sys.argv[1:]

    parser = build_args()
    args = parser.parse_args(args=argv)
    verbosity = args.verbose

    root = logging.getLogger()

    formatter = logging.Formatter('%(levelname).6s %(name)s %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    loglevels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    if verbosity >= len(loglevels):
        verbosity = len(loglevels) - 1

        level = loglevels[verbosity]
        root.setLevel(level)
        root.addHandler(handler)
    else:
        root.addHandler(logging.NullHandler())

    try:
        with open(args.script, "rb") as infile:
            binary_script = infile.read()
    except IOError as exc:
        print("ERROR: Unable to read script file: %s" % str(exc))
        return 1

    try:
        script = UpdateScript.FromBinary(binary_script, allow_unknown=args.allow_unknown, show_rpcs=args.show_rpcs)
    except ArgumentError as err:
        print("ERROR: ArgumentError: could not parse script")
        print(str(err))
        return 3
    except DataError as err:
        print("ERROR: DataError: could not parse script")
        print(str(err))
        return 4

    if args.format != 'text':
        print("ERROR: only text format is currently supported")
        return 2

    if args.format == 'text':
        print("\nUpdate Script")
        print("-------------")
        print("Path: %s" % args.script)
        print("Record Count: %d" % len(script.records))
        print("Total length: %d bytes" % len(binary_script))

        print("\nActions")
        print("-------")

        for i, record in enumerate(script.records):
            print("%02d: %s" % (i + 1, str(record)))

        print("")

    return 0



"""Auth providers are classes that can encrypt, decrypt, sign or verify data from IOTile devics

All auth providers must inherit from the AuthProvider class defined in this file and override the
methods that it provides with their own implementations.
"""

#pylint:disable=C0103

import pkg_resources
import hashlib
import struct
import hmac
from iotile.core.exceptions import NotFoundError, ArgumentError


class AuthProvider(object):
    """Base class for all objects that provide a way to authenticate or protect data.

    There are three kinds of things that are encrypted or signed in an IOTileDevice:

    1. Reports: Reports are signed using a unique per report key derived from a root
        key embedded on the device that guarantees its authenticity.
    2. RPCs: RPCs are signed using a per-session key derived from a token generated
        from a root key embedded on the device.
    3. Scripts: Scripts are signed using a unique key derived from a root key embedded
        on the device that guarantees its authenticity.

    Args:
        args (dict): An optional dictionary of AuthProvider specific config information
    """

    ReportKeyMagic = 0x00000002

    NoKey = 0
    UserKey = 1
    DeviceKey = 2

    KnownKeyRoots = {
        NoKey: 'no_key',
        UserKey: 'user_key',
        DeviceKey: 'device_key'
    }

    def __init__(self, args=None):
        if args is None:
            args = {}

        self.args = args

    @classmethod
    def FindByName(cls, name):
        """Find a specific installed auth provider by name."""

        for entry in pkg_resources.iter_entry_points('iotile.auth_provider'):
            if entry.name == name:
                return entry.load()

    @classmethod
    def VerifyRoot(cls, root):
        """Verify that the root key type is known to us.

        Raises:
            NotFoundError: If the key type is not known.
        """

        if root in cls.KnownKeyRoots:
            return

        raise NotFoundError("Unknown key type", key=root)

    @classmethod
    def DeriveReportKey(cls, root_key, report_id, sent_timestamp):
        """Derive a standard one time use report signing key.

        The standard method is HMAC-SHA256(root_key, MAGIC_NUMBER || report_id || sent_timestamp)
        where MAGIC_NUMBER is 0x00000002 and all integers are in little endian.
        """

        signed_data = struct.pack("<LLL", AuthProvider.ReportKeyMagic, report_id, sent_timestamp)

        hmac_calc = hmac.new(root_key, signed_data, hashlib.sha256)
        return bytearray(hmac_calc.digest())

    def encrypt_report(self, device_id, root, data, **kwargs):
        """Encrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should encrypt.
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        raise NotFoundError("encrypt_report method is not implemented")

    def decrypt_report(self, device_id, root, data, **kwargs):
        """Encrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should decrypt
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to decrypt the data.
        """

        raise NotFoundError("decrypt_report method is not implemented")

    def sign_report(self, device_id, root, data, **kwargs):
        """Sign a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should sign
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signature itself must always be a bytearray stored under the
                'signature' key, however additional keys may be present depending
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to sign the data.
        """

        raise NotFoundError("sign_report method is not implemented")

    def verify_report(self, device_id, root, data, signature, **kwargs):
        """Verify a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should verify
            signature (bytearray): The signature attached to data that we should verify
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to verify the data due to
                an error.  If the data is simply not valid, then the function returns
                normally.
        """

        raise NotFoundError("verify method is not implemented")



from .individual_format import IndividualReadingReport
from .report import IOTileReading, IOTileReport
from .signed_list_format import SignedListReport
from .broadcast import BroadcastReport
from .parser import IOTileReportParser
from .flexible_dictionary import FlexibleDictionaryReport
from .utc_assigner import UTCAssigner

__all__ = ['IndividualReadingReport', 'IOTileReport', 'IOTileReading',
           'BroadcastReport', 'SignedListReport', 'FlexibleDictionaryReport',
           'IOTileReportParser', 'UTCAssigner']



"""An ordered list of authentication providers that are checked in turn to attempt a crypto operation
"""

import pkg_resources
from iotile.core.exceptions import NotFoundError, ExternalError
from .auth_provider import AuthProvider


class ChainedAuthProvider(AuthProvider):
    """An AuthProvider that delegates operations to a list of subproviders.

    Each subprovider is checked in turn and the first one able to satisfy a request
    is used.  If no provider is available that can perform an operation, NotFoundError
    is raised.

    By default, the python entry_point group 'iotile.default_auth_providers' is used
    to build the list of auth providers in the chain.  The entries in that group should
    be tuples of (priority, auth_provider_class, arg_dict) where priority is an integer,
    auth_provider_class is an AuthProvider subclass and arg_dict is a dictionary of
    arguments passed to the constructor of auth_provider.
    """

    def __init__(self, args=None):
        super(ChainedAuthProvider, self).__init__(args)

        #FIXME: Allow overwriting default providers via args
        self._load_installed_providers()

        sub_providers = []
        for entry in pkg_resources.iter_entry_points('iotile.default_auth_providers'):
            priority, provider, args = entry.load()

            if provider not in self._auth_factories:
                raise ExternalError("Default authentication provider list references unknown auth provider", provider_name=provider, known_providers=self._auth_factories.keys())
            configured = self._auth_factories[provider](args)
            sub_providers.append((priority, configured))

        sub_providers.sort(key=lambda x: x[0])
        self.providers = sub_providers

    def _load_installed_providers(self):
        self._auth_factories = {}

        for entry in pkg_resources.iter_entry_points('iotile.auth_provider'):
            self._auth_factories[entry.name] = entry.load()

    def encrypt_report(self, device_id, root, data, **kwargs):
        """Encrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should encrypt.
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        for _priority, provider in self.providers:
            try:
                return provider.encrypt_report(device_id, root, data, **kwargs)
            except NotFoundError:
                pass

        raise NotFoundError("encrypt_report method is not implemented in any sub_providers")

    def decrypt_report(self, device_id, root, data, **kwargs):
        """Encrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should decrypt
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to decrypt the data.
        """

        for _priority, provider in self.providers:
            try:
                return provider.decrypt_report(device_id, root, data, **kwargs)
            except NotFoundError:
                pass

        raise NotFoundError("decrypt_report method is not implemented in any sub_providers")

    def sign_report(self, device_id, root, data, **kwargs):
        """Sign a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should sign
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signature itself must always be a bytearray stored under the
                'signature' key, however additional keys may be present depending
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to sign the data.
        """

        for _priority, provider in self.providers:
            try:
                return provider.sign_report(device_id, root, data, **kwargs)
            except NotFoundError:
                pass

        raise NotFoundError("sign_report method is not implemented in any sub_providers")

    def verify_report(self, device_id, root, data, signature, **kwargs):
        """Verify a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should verify
            signature (bytearray): The signature attached to data that we should verify
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to verify the data due to
                an error.  If the data is simply not valid, then the function returns
                normally.
        """

        for _priority, provider in self.providers:
            try:
                return provider.verify_report(device_id, root, data, signature, **kwargs)
            except NotFoundError:
                pass

        raise NotFoundError("verify_report method is not implemented in any sub_providers")



"""Entrypoint for virtual-device script that can serve a Virtual iotile device."""

from __future__ import unicode_literals, print_function, absolute_import
import argparse
import sys
import logging
import json
import imp
import os.path
import inspect
import pkg_resources
from future.utils import itervalues
from iotile.core.hw.virtual import VirtualIOTileDevice, VirtualIOTileInterface


def main(argv=None):
    """Serve access to a virtual IOTile device using a virtual iotile interface."""

    if argv is None:
        argv = sys.argv[1:]

    list_parser = argparse.ArgumentParser(add_help=False)
    list_parser.add_argument('-l', '--list', action='store_true', help="List all known installed interfaces and devices and then exit")

    parser = argparse.ArgumentParser(description="Serve acess to a virtual IOTile device using a virtual IOTile interface")

    parser.add_argument('interface', help="The name of the virtual device interface to use")
    parser.add_argument('device', help="The name of the virtual device to create")
    parser.add_argument('-v', '--verbose', action="count", default=0, help="Increase logging level (goes error, warn, info, debug)")
    parser.add_argument('-c', '--config', help="An optional JSON config file with arguments for the interface and device")
    parser.add_argument('-l', '--list', action='store_true', help="List all known installed interfaces and devices and then exit")
    parser.add_argument('-n', '--scenario', help="Load a test scenario from the given file")
    parser.add_argument('-s', '--state', help="Load a given state into the device before starting to serve it.  Only works with emulated devices.")
    parser.add_argument('-d', '--dump', help="Dump the device's state when we exit the program.  Only works with emulated devices.")
    parser.add_argument('-t', '--track', help="Track all changes to the device's state.  Only works with emulated devices.")

    args, _rest = list_parser.parse_known_args(argv)

    if args.list:
        #List out known virtual interfaces
        print("Installed Virtual Interfaces:")
        for entry in pkg_resources.iter_entry_points('iotile.virtual_interface'):
            print('- {}'.format(entry.name))

        print("\nInstalled Virtual Devices:")
        for entry in pkg_resources.iter_entry_points('iotile.virtual_device'):
            print('- {}'.format(entry.name))

        return 0

    args = parser.parse_args(argv)

    if args.verbose > 0:
        root = logging.getLogger()
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname).3s %(name)s %(message)s',
                                      '%y-%m-%d %H:%M:%S')
        handler.setFormatter(formatter)
        loglevels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
        if args.verbosity >= len(loglevels):
            args.verbosity = len(loglevels) - 1

        level = loglevels[args.verbosity]
        root.setLevel(level)
        root.addHandler(handler)

    config = {}
    iface = None
    if args.config is not None:
        with open(args.config, "r") as conf_file:
            config = json.load(conf_file)

    started = False
    device = None
    stop_immediately = args.interface == 'null'
    try:
        iface = instantiate_interface(args.interface, config)
        device = instantiate_device(args.device, config)

        if args.state is not None:
            print("Loading device state from file %s" % args.state)
            device.load_state(args.state)

        if args.scenario is not None:
            print("Loading scenario from file %s" % args.scenario)

            with open(args.scenario, "r") as infile:
                scenario = json.load(infile)

            # load_metascenario expects a list of scenarios even when there is only one
            if isinstance(scenario, dict):
                scenario = [scenario]

            device.load_metascenario(scenario)

        if args.track is not None:
            print("Tracking all state changes to device")
            device.state_history.enable()

        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter('%(asctime)s [AUDIT %(event_name)s] %(message)s'))

        iface.audit_logger.addHandler(handler)
        iface.audit_logger.setLevel(logging.INFO)

        iface.start(device)
        started = True

        print("Starting to serve virtual IOTile device")

        # We need to periodically process events that are queued up in the interface
        while True:
            iface.process()

            if stop_immediately:
                break

    except KeyboardInterrupt:
        print("Break received, cleanly exiting...")
    finally:
        if iface is not None and started:
            iface.stop()

        if args.dump is not None and device is not None:
            print("Dumping final device state to %s" % args.dump)
            device.save_state(args.dump)

        if args.track is not None and device is not None:
            print("Saving state history to file %s" % args.track)
            device.state_history.dump(args.track)

    return 0


def import_device_script(script_path):
    """Import a virtual device from a file rather than an installed module

    script_path must point to a python file ending in .py that contains exactly one
    VirtualIOTileDevice class definitions.  That class is loaded and executed as if it
    were installed.

    Args:
        script_path (string): The path to the script to load

    Returns:
        VirtualIOTileDevice: A subclass of VirtualIOTileDevice that was loaded from script_path
    """

    search_dir, filename = os.path.split(script_path)
    if search_dir == '':
        search_dir = './'

    if filename == '' or not os.path.exists(script_path):
        print("Could not find script to load virtual device, path was %s" % script_path)
        sys.exit(1)

    module_name, ext = os.path.splitext(filename)
    if ext != '.py':
        print("Script did not end with .py")
        sys.exit(1)

    try:
        file = None
        file, pathname, desc = imp.find_module(module_name, [search_dir])
        mod = imp.load_module(module_name, file, pathname, desc)
    finally:
        if file is not None:
            file.close()

    devs = [x for x in itervalues(mod.__dict__) if inspect.isclass(x) and issubclass(x, VirtualIOTileDevice) and x != VirtualIOTileDevice]
    if len(devs) == 0:
        print("No VirtualIOTileDevice subclasses were defined in script")
        sys.exit(1)
    elif len(devs) > 1:
        print("More than one VirtualIOTileDevice subclass was defined in script: %s" % str(devs))
        sys.exit(1)

    return devs[0]


def instantiate_device(virtual_dev, config):
    """Find a virtual device by name and instantiate it

    Args:
        virtual_dev (string): The name of the pkg_resources entry point corresponding to
            the device.  It should be in group iotile.virtual_device.  If virtual_dev ends
            in .py, it is interpreted as a python script and loaded directly from the script.
        config (dict): A dictionary with a 'device' key with the config info for configuring
            this virtual device.  This is optional.

    Returns:
        VirtualIOTileDevice: The instantiated subclass of VirtualIOTileDevice
    """
    conf = {}
    if 'device' in config:
        conf = config['device']

    #If we're given a path to a script, try to load and use that rather than search for an installed module
    if virtual_dev.endswith('.py'):
        dev = import_device_script(virtual_dev)
        return dev(conf)

    for entry in pkg_resources.iter_entry_points('iotile.virtual_device', name=virtual_dev):
        dev = entry.load()
        return dev(conf)

    print("Could not find an installed virtual device with the given name: {}".format(virtual_dev))
    sys.exit(1)


def instantiate_interface(virtual_iface, config):
    """Find a virtual interface by name and instantiate it

    Args:
        virtual_iface (string): The name of the pkg_resources entry point corresponding to
            the interface.  It should be in group iotile.virtual_interface
        config (dict): A dictionary with a 'interface' key with the config info for configuring
            this virtual interface.  This is optional.

    Returns:
        VirtualInterface: The instantiated subclass of VirtualInterface
    """

    # Allow the null virtual interface for testing
    if virtual_iface == 'null':
        return VirtualIOTileInterface()

    conf = {}
    if 'interface' in config:
        conf = config['interface']

    for entry in pkg_resources.iter_entry_points('iotile.virtual_interface', name=virtual_iface):
        interface = entry.load()
        return interface(conf)

    print("Could not find an installed virtual interface with the given name: {}".format(virtual_iface))
    sys.exit(1)



"""Update records that send an RPC."""

from __future__ import (print_function, absolute_import, unicode_literals)
from future.utils import python_2_unicode_compatible
from iotile.core.exceptions import ArgumentError
from ..record import UpdateRecord, MatchQuality


@python_2_unicode_compatible
class ResetDeviceRecord(UpdateRecord):
    """Reset the device.

    This record will cause the device to reset and then continue processing
    the next record in the script after it restarts.  This synchronous reset
    behavior allows scripts that produce side effects only visible after a
    reset but required for processing the rest of the script to trigger and
    wait for a reset before continuing.
    """

    RecordType = 5

    def encode_contents(self):
        """Encode the contents of this update record without including a record header.

        Returns:
            bytearary: The encoded contents.
        """

        return bytearray()

    @classmethod
    def MatchType(cls):
        """Return the record type that this record matches.

        All records must match an 8-bit record type field that is used to
        decode a binary script.  Note that multiple records may match the same
        8-bit record type if they have different levels of specificity.

        Returns:
            int: The single record type that this record matches.
        """

        return cls.RecordType

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this functon should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): The number of binary records that are included
                in record_data.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.
        """

        if record_count > 1:
            return MatchQuality.NoMatch

        # Allow other records to override us by matching against specific kinds of RPCs
        return MatchQuality.PerfectMatch

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This should be called with a binary record blob (NOT including the
        record type header) and it will decode it into a ResetDeviceRecord.

        Args:
            record_data (bytearray): The raw record data that we wish to parse
                into an UpdateRecord subclass NOT including its 8 byte record header.
            record_count (int): The number of records included in record_data.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.

        Returns:
            ResetDeviceRecord: The decoded reflash tile record.
        """

        if len(record_data) != 0:
            raise ArgumentError("Reset device record should have no included data", length=len(record_data))

        return ResetDeviceRecord()

    def __eq__(self, other):
        return isinstance(other, ResetDeviceRecord)

    def __str__(self):
        return "Reset device"



"""Commonly used utility functions shared throughout CoreTools."""

from .validating_dispatcher import ValidatingDispatcher
from .workqueue_thread import WorkQueueThread

__all__ = ['ValidatingDispatcher', 'WorkQueueThread']



"""Setup file for iotile-core."""

# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

# Caveats and possible issues
# Mac OS X
# - when using a virtualenv, readline is not properly installed into the virtualenv
#   and cannot be imported.  You need to install it using easy_install as described here
#   http://calvinx.com/tag/readline/


from setuptools import setup, find_packages
import version


setup(
    name="iotile-core",
    packages=find_packages(exclude=("test",)),
    version=version.version,
    license="LGPLv3",
    install_requires=[
        "ws4py>=0.3.5",
        "msgpack>=0.5.5",
        "python-dateutil>=2.6.0",
        "future>=0.16.0",
        "typedargs>=0.11.0",
        "sortedcontainers ~= 2.1",
        "monotonic ~= 1.5"
    ],
    extras_require={
        'ui': ["asciimatics>=1.9.0"]
    },
    entry_points={
        'console_scripts': [
            'iotile = iotile.core.scripts.iotile_script:main',
            'virtual_device = iotile.core.scripts.virtualdev_script:main',
            'iotile-updateinfo = iotile.core.scripts.iotile_updateinfo_script:main'
        ],
        'iotile.cmdstream': [
            'ws = iotile.core.hw.transport.websocketstream:WebSocketStream'
        ],
        'iotile.device_adapter': [
            'virtual = iotile.core.hw.transport.virtualadapter:VirtualDeviceAdapter'
        ],
        'iotile.report_format': [
            'individual = iotile.core.hw.reports:IndividualReadingReport',
            'signed_list = iotile.core.hw.reports:SignedListReport',
            'broadcast = iotile.core.hw.reports:BroadcastReport'
        ],
        'iotile.auth_provider': [
            'BasicAuthProvider = iotile.core.hw.auth.basic_auth_provider:BasicAuthProvider',
            'EnvAuthProvider = iotile.core.hw.auth.env_auth_provider:EnvAuthProvider',
            'ChainedAuthProvider = iotile.core.hw.auth.auth_chain:ChainedAuthProvider'
        ],
        'iotile.default_auth_providers': [
            'BasicAuthProvider = iotile.core.hw.auth.default_providers:DefaultBasicAuth',
            'EnvAuthProvider = iotile.core.hw.auth.default_providers:DefaultEnvAuth'
        ],
        'iotile.config_variables': [
            'iotile-core = iotile.core.config_variables:get_variables'
        ],
        'iotile.virtual_device': [
            'tile_based = iotile.core.hw.virtual.tile_based_device:TileBasedVirtualDevice'
        ],
        'iotile.recipe_action': [
            'FlashBoardStep = iotile.core.hw.debug.flash_board_step:FlashBoardStep'
        ],
        'iotile.app': [
            'device_info = iotile.core.hw.app.info_app',
            'device_updater = iotile.core.hw.app.updater_app'
        ]
    },
    description="IOTile Core Tools",
    author="Arch",
    author_email="info@arch-iot.com",
    url="https://github.com/iotile/coretools/iotilecore",
    keywords=["iotile", "arch", "embedded", "hardware"],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
IOTileCore
------

A python package for interacting with IOTile based devices.  See https://www.arch-iot.com.

"""
)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

import sys
from time import time
from iotile.core.utilities.typedargs import type_system

class ProgressBar:
    """
    A simple progress bar that updates itself in the console.

    If the program is being run in interactive mode, display the progress_bar
    otherwise do not display it
    """

    def __init__(self, title, count=100):
        self.title = title
        self.prog = 0 
        self.count = count


    def start(self):
        self.velocity = 0
        self.prog = 0
        self.start_time = time()

        if type_system.interactive:
            self.last_update = ""
            self._update()

    def _update(self, total=False):
        timestring = "Estimating time left"

        if self.velocity != 0.0:
            total_time = 1.0/self.velocity
            spent_time = self.current_time - self.start_time

            if total:
                time_left = spent_time
            else:
                time_left = total_time - spent_time

            units = "seconds"

            if time_left > 3600.:
                units = "hours"
                time_left /= 3600.0
            elif time_left > 60.0:
                units = "minutes"
                time_left /= 60.0

            comment = 'left'
            if total:
                comment = 'total'

            timestring = "%.1f %s %s" % (time_left, units, comment)

        if self.count > 0:
            x = int(self.prog * 40 // self.count)
        else:
            x = 0

        update_string = "%s: [%s%s] %s" % (self.title, '#'*x, '-'*(40-x), timestring)
        if len(update_string) < len(self.last_update):
            update_string += " "*(len(self.last_update) - len(update_string))


        #Erase the last update and write this one
        sys.stdout.write(chr(8)*len(self.last_update))
        sys.stdout.write(update_string)
        sys.stdout.flush()
        self.last_update = update_string

    def progress(self, x):
        self.current_time = time()

        if self.current_time > self.start_time and self.count > 0:
            self.velocity = (float(x)/self.count) / (self.current_time - self.start_time)

        self.prog = x

        if type_system.interactive:
            self._update()

    def end(self):
        self.prog = self.count
        if type_system.interactive:
            self._update(total=True)
            sys.stdout.write("\n")


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.
'''This file contains necessary functionality to manage the Hardware'''

from builtins import range
from functools import reduce
import time
import inspect
import os.path
import imp
import binascii
import sys
import logging
from queue import Empty
import pkg_resources
from future.utils import itervalues
from typedargs.annotate import annotated, param, return_type, finalizer, docannotate, context

from iotile.core.dev.semver import SemanticVersion
from iotile.core.hw.transport import CMDStream
from iotile.core.hw.exceptions import UnknownModuleTypeError
from iotile.core.exceptions import ArgumentError, HardwareError, ValidationError, TimeoutExpiredError, ExternalError
from iotile.core.dev.registry import ComponentRegistry
from iotile.core.hw.transport.adapterstream import AdapterCMDStream
from iotile.core.dev.config import ConfigManager
from iotile.core.hw.debug import DebugManager
from iotile.core.utilities.linebuffer_ui import LinebufferUI
from iotile.core.utilities.gid import uuid_to_slug

from .proxy import TileBusProxyObject
from .app import IOTileApp


@context("HardwareManager")
class HardwareManager(object):
    """
    A module for managing and interacting with IOTile Hardware

    This context provides tools to configure, control, debug and program
    any IOTile module.  Specific functionality can be implemented in dynamically
    loaded proxy objects that are designed to provide access to each IOTile.

    To create a HardwareManager, you need to pass a port string that describes the
    method to be used to connect to the IOTile device. The method should specify the
    name of the connection method optionally followed by a colon and any extra information
    possibly needed to connect using that method.

    Currently implemented ports are: (eventually auto-detect this)
        none
        bled112
        jlink
        jlink:mux=ftdi
        virtual:...(e.g. simple)

    Currently implemented devices are: (eventually auto-detect this too)
        nrf52
        lpc824

        if using the mux then append ;channel=[7..0] to the device
        (e.g.  --device="nrf52;channel=0")

    """

    # Allow overriding proxies for development by adding them to this shared proxy map
    DevelopmentProxies = {}
    DevelopmentApps = {}
    DevelopmentAppNames = {}

    logger = logging.getLogger(__name__)

    @param("port", "string", desc="transport method to use in the format transport[:port[,connection_string]]")
    @param("record", "path", desc="Optional file to record all RPC calls and responses made on this HardwareManager")
    def __init__(self, port=None, record=None, adapter=None):
        if port is None and adapter is None:
            try:
                conf = ConfigManager()
                port = conf.get('core:default-port')
            except ArgumentError:
                raise ArgumentError("No port given and no core:default-port config variable set", suggestion="Specify the port to use to connect to the IOTile devices")
        elif port is None:
            port = ""

        transport, _, arg = port.partition(':')

        self.transport = transport
        self.port = None
        if arg != "":
            self.port = arg

        self._record = record

        self.stream = self._create_stream(adapter)

        self._stream_queue = None
        self._trace_queue = None
        self._broadcast_queue = None
        self._trace_data = bytearray()

        self._proxies = {'TileBusProxyObject': TileBusProxyObject}
        self._name_map = {TileBusProxyObject.ModuleName(): [TileBusProxyObject]}

        self._known_apps = {}
        self._named_apps = {}

        self._setup_proxies()
        self._setup_apps()

    @classmethod
    def RegisterDevelopmentProxy(cls, proxy_obj):  # pylint: disable=C0103; class methods are capitalized when expected to be invoked on types
        """
        Register a proxy object that should be available for local development.

        Often during development, you need to create a virtual iotile device with
        its own proxy object.  It's easy to point CoreTools at the virtual device
        in order to test it, but there did not use to be a good way to load in
        its proxy object.  This method allows the user to inject a development
        proxy module to use with the virtual device.

        Args:
            proxy_obj (TileBusProxyObject): The proxy module class that should be
                registered.
        """

        name = proxy_obj.ModuleName()

        if name not in HardwareManager.DevelopmentProxies:
            HardwareManager.DevelopmentProxies[name] = []

        HardwareManager.DevelopmentProxies[name].append(proxy_obj)

    @classmethod
    def ClearDevelopmentApps(cls):
        """Clear all development apps previously registered."""

        cls.DevelopmentAppNames = {}
        cls.DevelopmentApps = {}

    @classmethod
    def RegisterDevelopmentApp(cls, app):  # pylint: disable=C0103; class methods are capitalized when expected to be invoked on types
        """Register an IOTileApp object that should be available for local development.

        Often during development, you need to create a virtual iotile device with
        its own app object.  It's easy to point CoreTools at the virtual device
        in order to test it, but there did not use to be a good way to load in
        its app object.  This method allows the user to inject a development
        app module to use with the virtual device.

        Args:
            app (TileBusApp): The app module class that should be
                registered.
        """

        matches = app.MatchInfo()

        for (app_tag, ver_range, quality) in matches:
            if app_tag not in HardwareManager.DevelopmentApps:
                HardwareManager.DevelopmentApps[app_tag] = []

            HardwareManager.DevelopmentApps[app_tag].append((ver_range, quality, app))

        HardwareManager.DevelopmentAppNames[app.AppName()] = app

    def load_development_proxy(self, filename):
        """
        Load a proxy object instead of having it pre-registered.
        This loads the proxy for a python file that we want to use in test fixtures

        This is useful for pulling in the TileBusProxyObject-related classes for a test
        so that they don't need to be tethered to the HardwareManager (i.e. VirtualTile)


        Args:
            filename (str): The proxy module class that should be
                loaded.
        """

        proxy_class = self._load_module_classes(filename, TileBusProxyObject)[0]

        name = proxy_class.ModuleName()
        if name not in HardwareManager.DevelopmentProxies:
            HardwareManager.DevelopmentProxies[name] = []

        HardwareManager.DevelopmentProxies[name].append(proxy_class)

    def _setup_proxies(self):
        """Load in proxy module objects for all of the registered components on this system."""

        # Find all of the registered IOTile components and see if we need to add any proxies for them
        reg = ComponentRegistry()
        modules = reg.list_components()

        proxies = reduce(lambda x, y: x+y, [reg.find_component(x).proxy_modules() for x in modules], [])
        proxy_classes = []
        for prox in proxies:
            proxy_classes += self._load_module_classes(prox, TileBusProxyObject)

        # Find all installed proxy objects through registered entry points
        for entry in pkg_resources.iter_entry_points('iotile.proxy'):
            mod = entry.load()
            proxy_classes += [x for x in itervalues(mod.__dict__) if inspect.isclass(x) and issubclass(x, TileBusProxyObject) and x != TileBusProxyObject]

        for obj in proxy_classes:
            if obj.__name__ in self._proxies:
                continue #Don't readd proxies that we already know about

            self._proxies[obj.__name__] = obj

            #Check if this object matches a specific shortened name so that we can
            #automatically match a hw module to a proxy without user intervention
            try:
                short_name = obj.ModuleName()
                if short_name in self._name_map:
                    self._name_map[short_name].append(obj)
                else:
                    self._name_map[short_name] = [obj]
            except Exception:  #pylint: disable=broad-except;We don't want this to die if someone loads a misbehaving plugin
                self.logger.exception("Error importing misbehaving proxy module, skipping.")

    def _setup_apps(self):
        """Load in all iotile app objects for all registered or installed components on this system."""

        reg = ComponentRegistry()
        modules = reg.list_components()

        apps = reduce(lambda x, y: x+y, [reg.find_component(x).app_modules() for x in modules], [])
        app_classes = []
        for app in apps:
            app_classes += self._load_module_classes(app, IOTileApp)

        # Find all installed proxy objects through registered entry points
        for entry in pkg_resources.iter_entry_points('iotile.app'):
            mod = entry.load()
            app_classes += [x for x in itervalues(mod.__dict__) if inspect.isclass(x) and issubclass(x, IOTileApp) and x != IOTileApp]

        for app in app_classes:
            try:
                matches = app.MatchInfo()
                name = app.AppName()
                for tag, ver_range, quality in matches:
                    if tag not in self._known_apps:
                        self._known_apps[tag] = []

                    self._known_apps[tag].append((ver_range, quality, app))

                if name in self._named_apps:
                    self.logger.warning("Added an app module with an existing name, overriding previous app, name=%s", name)

                self._named_apps[name] = app
            except Exception:  #pylint: disable=broad-except;We don't want this to die if someone loads a misbehaving plugin
                self.logger.exception("Error importing misbehaving app module, skipping.")

    @param("address", "integer", "positive", desc="numerical address of module to get")
    @param("basic", "bool", desc="return a basic global proxy rather than a specialized one")
    def get(self, address, basic=False):
        """Create a proxy object for a tile by address.

        The correct proxy object is determined by asking the tile for its
        status information and looking up the appropriate proxy in our list of
        installed proxy objects.  If you want to send raw RPCs, you can get a
        basic TileBusProxyObject by passing basic=True.
        """

        tile = self._create_proxy('TileBusProxyObject', address)

        if basic:
            return tile

        name = tile.tile_name()
        version = tile.tile_version()

        # Now create the appropriate proxy object based on the name and version of the tile
        tile_type = self.get_proxy(name)
        if tile_type is None:
            raise HardwareError("Could not find proxy object for tile", name="'{}'".format(name), known_names=self._name_map.keys())

        tile = tile_type(self.stream, address)
        tile._hwmanager = self

        return tile

    @docannotate
    def app(self, name=None, path=None):
        """Find the best IOTileApp for the device we are connected to.

        Apps are matched by looking at the app tag and version information
        specified by the connected device.  If no installed app matches, an
        exception will be thrown.  You can also force the matching of a
        specific app by using the name parameter.

        Args:
            name (str): Optional name of the app that you wish to load.
            path (str): Optional path to a python file containing the
                app that you wish to load.

        Returns:
            IOTileApp show-as context: The IOTileApp class that was loaded
                for this device.
        """

        if name is not None and path is not None:
            raise ArgumentError("You cannot specify both an app name and an app path", name=name, path=path)

        # We perform all app matching by asking the device's controller for its app and os info
        tile = self._create_proxy('TileBusProxyObject', 8)
        device_id, os_info, app_info = tile.rpc(0x10, 0x08, result_format="L8xLL")

        os_tag = os_info  & ((1 << 20) - 1)
        os_version_str = '%d.%d.%d'  % ((os_info >> 26) & ((1 << 6) - 1), (os_info >> 20) & ((1 << 6) - 1), 0)

        app_tag = app_info & ((1 << 20) - 1)
        app_version_str = '%d.%d.%d'  % ((app_info >> 26) &  ((1 << 6) - 1), (app_info>>20) & ((1 << 6) - 1), 0)

        os_version = SemanticVersion.FromString(os_version_str)
        app_version = SemanticVersion.FromString(app_version_str)

        app_class = None

        # If name includes a .py, assume that it points to python file and try to load that.
        if name is None and path is not None:
            loaded_classes = self._load_module_classes(path, IOTileApp)
            if len(loaded_classes) > 1:
                raise ArgumentError("app called with a python file that contained more than one IOTileApp class", classes=loaded_classes)
            elif len(loaded_classes) == 0:
                raise ArgumentError("app called with a python file that did not contain any IOTileApp subclasses")

            app_class = loaded_classes[0]
        elif name is not None:
            if name in self.DevelopmentAppNames:
                app_class = self.DevelopmentAppNames[name]
            else:
                app_class = self._named_apps.get(name)
        else:
            best_match = None
            matching_tags = self._known_apps.get(app_tag, [])
            dev_tags = self.DevelopmentApps.get(app_tag, [])

            for (ver_range, quality, app) in matching_tags + dev_tags:
                if ver_range.check(app_version):
                    if best_match is None:
                        best_match = (quality, app)
                    elif quality > best_match[0]:
                        best_match = (quality, app)

            if best_match is not None:
                app_class = best_match[1]

        if app_class is None:
            raise HardwareError("Could not find matching application for device", app_tag=app_tag, explicit_app=name, installed_apps=[x for x in self._named_apps])

        app = app_class(self, (app_tag, app_version), (os_tag, os_version), device_id)
        return app

    @annotated
    def controller(self):
        """
        Find an attached IOTile controller and attempt to connect to it.
        """

        con = self.get(8)
        con._hwmanager = self

        return con

    @param("device_uuid", "integer", desc="UUID of the device we would like to connect to")
    @param("wait", "float", desc="Time to wait for devices to show up before connecting")
    def connect(self, device_uuid, wait=None):
        """Attempt to connect to a device by its UUID
        """

        self.stream.connect(device_uuid, wait=wait)

    @param("connection_string", "string", desc="opaque connection string indicating which device")
    def connect_direct(self, connection_string):
        """Attempt to connect to a device using a connection string
        """

        self.stream.connect_direct(connection_string)

    @annotated
    def disconnect(self):
        """Attempt to disconnect from a device
        """

        self._trace_queue = None
        self._stream_queue = None

        self.stream.disconnect()

    @param("connection_string", "string", desc="opaque connection string indicating which device")
    def debug(self, connection_string=None):
        """Prepare the device for debugging if supported.

        Some transport mechanisms support a low level debug channel
        that permits recovery and test operations such as erasing
        and forcibly reprogramming a device or dumping memory.

        No debug operations are supported, this function will raise
        an exception.
        """

        self.stream.enable_debug(connection_string)
        return DebugManager(self.stream)

    @return_type("bool")
    def heartbeat(self):
        """
        Check whether the underlying command stream is functional
        """

        return self.stream.heartbeat()

    @annotated
    def enable_broadcasting(self):
        """Enable the collection of broadcast IOTile reports.

        Broadcast reports contain tagged readings from an IOTile device
        but are sent without a connection to that device.  The specific
        method that is used to broadcast the report varies by connection
        technology but it could be, e.g., a bluetooth low energy advertising
        packet.

        By default all broadcast reports are dropped unless you call
        enable_broadcasting to allocate a queue to receive them.

        There does not need to be an active connection for you to call
        enable_broadcasting.

        Once you call enable_broadcasting, it remains in effect for the
        duration of this HardwareManager object.
        """

        if self._broadcast_queue is not None:
            return

        self._broadcast_queue = self.stream.enable_broadcasting()

    @annotated
    def enable_streaming(self):
        """Enable streaming of report data from the connected device.

        This function will create an internal queue to receive and store
        reports until the user looks at them and then inform the connected
        IOTile device that is should begin streaming data.

        This is done by telling the underlying DeviceAdapter managing the
        communication with the device that it should open the device's
        streaming interface.

        There is currently no way to close the streaming interface except
        by disconnecting from the device and then reconnecting to it.
        """

        if self._stream_queue is not None:
            return

        self._stream_queue = self.stream.enable_streaming()

    @annotated
    def enable_tracing(self):
        """Enable tracing of realtime debug information over this interface"""

        if self._trace_queue is not None:
            return

        self._trace_queue = self.stream.enable_tracing()

    @return_type("integer")
    def count_reports(self):
        """Return the current size of the reports queue"""
        if self._stream_queue is None:
            return 0

        return self._stream_queue.qsize()

    @docannotate
    def watch_broadcasts(self, whitelist=None, blacklist=None):
        """Spawn an interactive terminal UI to watch broadcast data from devices.

        Devices are allowed to post a broadcast report containing stream data.
        This function will create a list in your console window with the latest
        broadcast value from each device in range.

        Args:
            whitelist (list(integer)): Only include devices with these listed ids.
            blacklist (list(integer)): Include every device **except** those with these
                specific ids.  If combined with whitelist, whitelist wins and this
                parameter has no effect.
        """

        if whitelist is not None:
            whitelist = set(whitelist)

        if blacklist is not None:
            blacklist = set(blacklist)

        def _title(_items):
            return ["Watching Broadcast Reports (Ctrl-C to Stop)"]

        def _poll():
            results = [x for x in self.iter_broadcast_reports(blocking=False)]
            return results

        def _text(item):
            fmt_uuid = "%08X" % item.origin
            fmt_uuid = fmt_uuid[:4] + '-' + fmt_uuid[4:]

            reading = item.visible_readings[0]
            return "{0: <15} stream: {1: 04X}    value: {2: <8}".format(fmt_uuid, reading.stream, reading.value)

        def _sort_order(item):
            return item.origin

        def _hash(item):
            uuid = item.origin
            if whitelist is not None and uuid not in whitelist:
                return None

            if blacklist is not None and whitelist is None and uuid in blacklist:
                return None

            return uuid

        line_ui = LinebufferUI(_poll, _hash, _text, sortkey_func=_sort_order, title=_title)
        line_ui.run()

    @docannotate
    def watch_scan(self, whitelist=None, blacklist=None, sort="id"):
        """Spawn an interactive terminal UI to watch scan results.

        This is just a fancy way of calling scan() repeatedly and
        deduplicating results per device so that each one has a static place
        on the screen.

        You can decide how you want to order the results with the sort parameter.

        If you pick "id", the default, then results will have a largely static
        order based on the UUID of each device so that there will not be too
        much screen churn.

        Args:
            whitelist (list(integer)): Only include devices with these listed ids.
            blacklist (list(integer)): Include every device **except** those with these
                specific ids.  If combined with whitelist, whitelist wins and this
                parameter has no effect.
            sort (str): The specific way to sort the list on the screen.  Options are
                id, signal.  Defaults to id.
        """

        if whitelist is not None:
            whitelist = set(whitelist)

        if blacklist is not None:
            blacklist = set(blacklist)

        def _title(items):
            return ["Realtime Scan: %d Devices in Range" % len(items)]

        def _poll():
            return self.scan()

        def _text(item):
            fmt_uuid = "%08X" % item['uuid']
            fmt_uuid = fmt_uuid[:4] + '-' + fmt_uuid[4:]

            return "{0: <15} signal: {1: <7d} connected: {2: <8}".format(fmt_uuid, item['signal_strength'], str(item.get('user_connected', 'unk')))

        def _sort_order(item):
            if sort == "signal":
                return -item['signal_strength']

            return item['uuid']

        def _hash(item):
            uuid = item['uuid']
            if whitelist is not None and uuid not in whitelist:
                return None

            if blacklist is not None and whitelist is None and uuid in blacklist:
                return None

            return uuid

        line_ui = LinebufferUI(_poll, _hash, _text, sortkey_func=_sort_order, title=_title)
        line_ui.run()

    @docannotate
    def watch_reports(self, whitelist=None, blacklist=None):
        """Spawn an interactive terminal UI to watch reports once connected to a device.

        Args:
            whitelist (list(integer)): Only include streams with these listed ids.
            blacklist (list(integer)): Include every stream **except** those with these
                specific ids.  If combined with whitelist, whitelist wins and this
                parameter has no effect.
        """

        if whitelist is not None:
            whitelist = set(whitelist)

        if blacklist is not None:
            blacklist = set(blacklist)

        def _title(items):
            base = "Watching Report for Device ID "
            if items:
                base = base + str(items[list(items.keys())[0]].object.origin)
            meta = "{:15s} {:4s}  {:8s}".format("Last Timestamp", "Stream ID", "Stream Value")
            return [base, meta]

        def _poll():
            results = [x for x in self.iter_reports(blocking=False)]
            return results

        def _text(item):
            reading = item.visible_readings[0]
            return "{0:<15} {1:04X}         value: {2:<8}".format(reading.raw_time, reading.stream, reading.value)

        def _sort_order(item):
            return item.origin

        def _hash(item):
            stream = item.visible_readings[0].stream

            if whitelist is not None and stream not in whitelist:
                return None

            if blacklist is not None and whitelist is None and stream in blacklist:
                return None

            return stream

        if not self.stream.connected:
            print("Not connected to a device. Please connect first")
            return

        if not self._stream_queue:
            print("Enable streaming to watch reports")
            return

        line_ui = LinebufferUI(_poll, _hash, _text, sortkey_func=_sort_order, title=_title)
        line_ui.run()


    @return_type("string")
    @param("encoding", "string", desc="The encoding to use to dump the trace, either 'hex' or 'raw'")
    def dump_trace(self, encoding):
        """Dump all received tracing data currently received from the device to stdout

        The data is encoded per the encoding parmeter which must be either
        the string 'hex' or 'raw'.  If hex is passed, the data is printed as hex digits,
        if raw is passed, the data is printed as received from the device.
        """

        if encoding not in ['raw', 'hex']:
            raise ValidationError("Unknown encoding type specified in dump trace", encoding=encoding, known_encodings=['hex', 'raw'])

        if self._trace_queue is None:
            return ""

        self._accumulate_trace()

        if encoding == 'raw':
            return bytes(self._trace_data)

        return binascii.hexlify(self._trace_data).decode('utf-8')

    def wait_trace(self, size, timeout=None, drop_before=False, progress_callback=None):
        """Wait for a specific amount of tracing data to be received.

        This function is the equivalent of wait_reports for streaming data.
        It allows you to block until a specific amount of tracing data has
        been received.  The optional timeout parameter allows you to stop
        waiting if you never receive enough tracing data after a specific
        amount of time.

        Args:
            size (int): The number of bytes to wait for.
            timeout (float): The maximum number of seconds to wait for
                these bytes to be received.
            drop_before (bool): Truncate all data received until now
                before waiting for size bytes.
            progress_callback (callable): An optional progress callback that
                is called periodically with updates.  It should have the
                signature progress_callback(received_byte_count, total_byte_count)

        Returns:
            bytearray: The raw trace data obtained.
        """

        if drop_before:
            self._trace_data = bytearray()

        if progress_callback is None:
            progress_callback = lambda x, y: None

        if len(self._trace_data) >= size:
            progress_callback(size, size)

            data = self._trace_data[:size]
            self._trace_data = self._trace_data[size:]

            return data

        progress_callback(len(self._trace_data), size)

        start = time.time()
        while len(self._trace_data) < size:
            progress_callback(len(self._trace_data), size)
            self._accumulate_trace()

            time.sleep(0.1)
            now = time.time()

            if timeout is not None and ((now - start) > timeout):
                raise TimeoutExpiredError("Timeout waiting for tracing data", expected_size=size, received_size=len(self._trace_data), timeout=timeout)

        progress_callback(size, size)

        data = self._trace_data[:size]
        self._trace_data = self._trace_data[size:]

        return data

    def _accumulate_trace(self):
        """Copy tracing data from trace queue into _trace_data
        """

        if self._trace_queue is None:
            return

        try:
            while True:
                blob = self._trace_queue.get(block=False)
                self._trace_data += bytearray(blob)
        except Empty:
            pass

    def iter_broadcast_reports(self, blocking=False):
        """Iterate over broadcast reports that have been received.

        This function is designed to allow the creation of dispatch or
        processing systems that process broadcast reports as they come in.

        Args:
            blocking (bool): Whether to stop when there are no more readings or
                block and wait for more.
        """

        if self._broadcast_queue is None:
            return

        try:
            while True:
                yield self._broadcast_queue.get(block=blocking)
        except Empty:
            pass

    def wait_broadcast_reports(self, num_reports, timeout=2.0):
        """Wait until a specific number of broadcast reports have been received.

        Args:
            num_reports (int): The number of reports to wait for
            timeout (float): The maximum number of seconds to wait without
                receiving another report.
        """

        if self._broadcast_queue is None:
            raise ExternalError("You have to enable broadcasting before you can wait for broadcast reports")

        reports = []

        for i in range(0, num_reports):
            try:
                report = self._broadcast_queue.get(timeout=timeout)
                reports.append(report)
            except Empty:
                raise TimeoutExpiredError("Timeout waiting for a report", expected_number=num_reports, received_number=i, received_reports=reports)

        return reports



    def iter_reports(self, blocking=False):
        """Iterate over reports that have been received.

        If blocking is True, this iterator will never stop.  Otherwise
        it will iterate over all reports currently in the queue (and those
        added during iteration)

        Args:
            blocking (bool): Whether to stop when there are no more readings or
                block and wait for more.
        """
        if self._stream_queue is None:
            return

        try:
            while True:
                yield self._stream_queue.get(block=blocking)
        except Empty:
            pass

    def wait_reports(self, num_reports, timeout=2.0):
        """Wait for a fixed number of reports to be received

        Args:
            num_reports (int): The number of reports to wait for
            timeout (float): The maximum number of seconds to wait without
                receiving another report.
        """

        if self._stream_queue is None:
            raise ExternalError("You have to enable streaming before you can wait for reports")

        reports = []

        for i in range(0, num_reports):
            try:
                report = self._stream_queue.get(timeout=timeout)
                reports.append(report)
            except Empty:
                raise TimeoutExpiredError("Timeout waiting for a report", expected_number=num_reports, received_number=i, received_reports=reports)

        return reports

    @annotated
    def reset(self):
        """
        Attempt to reset the underlying stream back to a known state
        """

        self.stream.reset()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stream.close()
        return False

    @finalizer
    def close(self):
        """Close the current AdapterStream"""
        self.stream.close()

    @classmethod
    def _load_module_classes(cls, path, base_class):
        """Load a python module and return all classes that inherit from a given base."""

        folder, basename = os.path.split(path)
        basename, ext = os.path.splitext(basename)
        if ext not in (".py", ".pyc", ""):
            raise ArgumentError("Attempted to load module is not a python package or module (.py or .pyc)", path=path)

        try:
            fileobj, pathname, description = imp.find_module(basename, [folder])

            #Don't load modules twice
            if basename in sys.modules:
                mod = sys.modules[basename]
            else:
                mod = imp.load_module(basename, fileobj, pathname, description)
        except ImportError as exc:
            cls.logger.exception("Error importing module: %s looking for class %s", path, base_class)
            raise ArgumentError("Could not import module in order to load external proxy modules", module_path=path, parent_directory=folder, module_name=basename, error=str(exc))

        # Find all classes in this module that inherit from the given base class
        return [x for x in itervalues(mod.__dict__) if inspect.isclass(x) and issubclass(x, base_class) and x != base_class]

    @return_type("list(basic_dict)")
    @param("wait", "float", desc="Time to wait for devices to show up before returning")
    @param("sort", "string", desc="Sort scan results by a key named key")
    @param("limit", "integer", desc="Limit results to the first n devices")
    @param("reverse", "bool", desc="Reverse the sort order")
    def scan(self, wait=None, sort=None, reverse=False, limit=None):
        """Scan for available devices and print a dictionary of information about them.

        If wait is specified as a floating point number in seconds, then the default wait times
        configured inside of the stream or device adapter used to find IOTile devices is overriden
        with the value specified.

        Args:
            wait (float): An optional override time to wait for results to accumulate before returning
        """

        devices = self.stream.scan(wait=wait)

        for device in devices:
            # Add a Device Slug for user convenience
            if 'uuid' in device:
                device['slug'] = uuid_to_slug(device['uuid'])

        if sort is not None:
            devices.sort(key=lambda x: x[sort], reverse=reverse)

        if limit is not None:
            devices = devices[:limit]

        # FIXME: Use dictionary format in bled112stream to document information returned about devices
        return devices

    def get_proxy(self, short_name):
        """
        Find a proxy type given its short name.

        If no proxy type is found, return None.
        """

        if short_name in HardwareManager.DevelopmentProxies:
            return HardwareManager.DevelopmentProxies[short_name][0]

        if short_name not in self._name_map:
            return None

        return self._name_map[short_name][0]

    def _create_proxy(self, proxy, address):
        """
        Create a python proxy object to talk to an IOTile module with the given type
        at the given address.
        """

        if proxy not in self._proxies:
            raise UnknownModuleTypeError("unknown proxy module specified", module_type=proxy, known_types=self._proxies.keys())

        proxy_class = self._proxies[proxy]
        return proxy_class(self.stream, address)

    def _create_stream(self, force_adapter=None):
        conn_string = None
        port = self.port
        if port is not None and "," in port:
            port, conn_string = port.split(',')

        if port is not None:
            port = port.strip()
        if conn_string is not None:
            conn_string = conn_string.strip()

        #Check if we're supposed to use a specific device adapter
        if force_adapter is not None:
            return AdapterCMDStream(force_adapter, port, conn_string, record=self._record)

        #First check if this is the special none stream that creates a transport channel nowhere
        if self.transport == 'none':
            return CMDStream(port, conn_string, record=self._record)

        #Next attempt to find a CMDStream that is registered for this transport type
        for stream_entry in pkg_resources.iter_entry_points('iotile.cmdstream'):
            if stream_entry.name != self.transport:
                continue

            stream_factory = stream_entry.load()
            return stream_factory(port, conn_string, record=self._record)

        #Otherwise attempt to find a DeviceAdapter that we can turn into a CMDStream
        for adapter_entry in pkg_resources.iter_entry_points('iotile.device_adapter'):
            if adapter_entry.name != self.transport:
                continue

            adapter_factory = adapter_entry.load()
            return AdapterCMDStream(adapter_factory(port), port, conn_string, record=self._record)

        raise HardwareError("Could not find transport object registered to handle passed transport type", transport=self.transport)
